from typing import TYPE_CHECKING

import pytest

from expanse.core.application import Application
from expanse.core.http.gateway import Gateway
from expanse.http.middleware.manage_cors import ManageCors
from expanse.http.response import Response
from expanse.routing.responder import Responder
from expanse.testing.client import TestClient


if TYPE_CHECKING:
    from expanse.routing.router import Router


@pytest.fixture(autouse=True)
def setup_app(app: Application) -> None:
    gateway = app.container.get(Gateway)
    gateway.prepend_middleware(ManageCors)

    router: Router = app.container.get("router")

    def ping(responder: Responder) -> Response:
        return responder.json("pong")

    def error(responder: Responder) -> Response:
        return responder.abort(500)

    router.post("/api/ping", ping)
    router.post("/api/error", error)
    router.post("/web/ping", lambda: Response("pong", content_type="text/plain"))

    app.config["cors"] = {
        "paths": ["api/*"],
        "supports_credentials": False,
        "allowed_origins": ["http://localhost"],
        "allowed_headers": ["X-Custom-1", "X-Custom-2"],
        "allowed_methods": ["GET", "POST"],
        "exposed_headers": [],
        "max_age": 0,
    }


def test_it_should_return_access_control_allow_origin_when_no_origin_on_request(
    client: TestClient,
) -> None:
    response = client.options(
        "/api/ping", headers={"Access-Control-Request-Method": "POST"}
    )

    assert response.status_code == 204
    assert response.headers["Access-Control-Allow-Origin"] == "http://localhost"


def test_it_should_return_access_control_allow_origin_with_origin_specified(
    client: TestClient,
) -> None:
    response = client.options(
        "/api/ping",
        headers={"Origin": "http://localhost", "Access-Control-Request-Method": "POST"},
    )

    assert response.status_code == 204
    assert response.headers["Access-Control-Allow-Origin"] == "http://localhost"


def test_allow_all_origins(
    client: TestClient,
) -> None:
    client.app.config["cors"]["allowed_origins"] = ["*"]
    response = client.options(
        "/api/ping",
        headers={"Origin": "http://localhost", "Access-Control-Request-Method": "POST"},
    )

    assert response.status_code == 204
    assert response.headers["Access-Control-Allow-Origin"] == "*"


def test_allow_exact_origin(
    client: TestClient,
) -> None:
    client.app.config["cors"]["allowed_origins"] = [
        "http://localhost",
        "http://localhost2",
    ]
    response = client.options(
        "/api/ping",
        headers={
            "Origin": "http://localhost2",
            "Access-Control-Request-Method": "POST",
        },
    )

    assert response.status_code == 204
    assert response.headers["Access-Control-Allow-Origin"] == "http://localhost2"


def test_allow_all_origins_wildcard(
    client: TestClient,
) -> None:
    client.app.config["cors"]["allowed_origins"] = ["*.python-expanse.org"]
    response = client.options(
        "/api/ping",
        headers={
            "Origin": "http://test.python-expanse.org",
            "Access-Control-Request-Method": "POST",
        },
    )

    assert response.status_code == 204
    assert (
        response.headers["Access-Control-Allow-Origin"]
        == "http://test.python-expanse.org"
    )


def test_allowed_origins_wildcard_includes_nested_subdomain(
    client: TestClient,
) -> None:
    client.app.config["cors"]["allowed_origins"] = ["*.python-expanse.org"]
    response = client.options(
        "/api/ping",
        headers={
            "Origin": "http://foo.bar.test.python-expanse.org",
            "Access-Control-Request-Method": "POST",
        },
    )

    assert response.status_code == 204
    assert (
        response.headers["Access-Control-Allow-Origin"]
        == "http://foo.bar.test.python-expanse.org"
    )


def test_allowed_origins_wildcard_no_match(
    client: TestClient,
) -> None:
    client.app.config["cors"]["allowed_origins"] = ["*.python-expanse.org"]
    response = client.options(
        "/api/ping",
        headers={
            "Origin": "http://test.python-poetry.org",
            "Access-Control-Request-Method": "POST",
        },
    )

    assert response.status_code == 204
    assert "Access-Control-Allow-Origin" not in response.headers


def test_options_allowed_origins_non_existing_route(
    client: TestClient,
) -> None:
    response = client.options(
        "/api/missing",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Method": "POST",
        },
    )

    assert response.status_code == 204
    assert response.headers["Access-Control-Allow-Origin"] == "http://localhost"


def test_options_origin_not_allowed(
    client: TestClient,
) -> None:
    response = client.options(
        "/api/ping",
        headers={
            "Origin": "http://notlocalhost",
            "Access-Control-Request-Method": "POST",
        },
    )

    assert response.status_code == 204
    assert response.headers["Access-Control-Allow-Origin"] == "http://localhost"


def test_method_allowed(
    client: TestClient,
) -> None:
    response = client.post(
        "/api/ping",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Method": "POST",
        },
    )

    assert response.status_code == 200
    assert "Access-Control-Allow-Methods" not in response.headers
    assert response.json() == "pong"


def test_method_not_allowed(
    client: TestClient,
) -> None:
    response = client.post(
        "/api/ping",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Method": "PUT",
        },
    )

    assert response.status_code == 200
    assert "Access-Control-Allow-Methods" not in response.headers


def test_method_allow_all_methods_allowed(
    client: TestClient,
) -> None:
    client.app.config["cors"]["allowed_methods"] = ["*"]

    response = client.options(
        "/api/ping",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Method": "PUT",
        },
    )

    assert response.status_code == 204
    assert response.headers["Access-Control-Allow-Methods"] == "PUT"


def test_allowed_headers_allow_options(
    client: TestClient,
) -> None:
    response = client.options(
        "/api/ping",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Method": "POST",
            "Access-Control-Request-Headers": "X-Custom-1, X-Custom-2",
        },
    )

    assert response.status_code == 204
    assert (
        response.headers["Access-Control-Allow-Headers"]
        == "X-Custom-1, X-Custom-2".lower()
    )
    assert response.text == ""


def test_allowed_headers_allow_wildcard(
    client: TestClient,
) -> None:
    client.app.config["cors"]["allowed_headers"] = ["*"]

    response = client.options(
        "/api/ping",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Method": "POST",
            "Access-Control-Request-Headers": "X-Custom-3",
        },
    )

    assert response.status_code == 204
    assert response.headers["Access-Control-Allow-Headers"] == "X-Custom-3"
    assert response.text == ""


def test_allowed_headers_not_allowed(
    client: TestClient,
) -> None:
    response = client.options(
        "/api/ping",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Method": "POST",
            "Access-Control-Request-Headers": "X-Custom-3",
        },
    )

    assert response.status_code == 204
    assert (
        response.headers["Access-Control-Allow-Headers"]
        == "X-Custom-1, X-Custom-2".lower()
    )
    assert response.text == ""


def test_allowed_headers_allowed(
    client: TestClient,
) -> None:
    response = client.post(
        "/api/ping",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Headers": "X-Custom-1, X-Custom-2",
        },
    )

    assert response.status_code == 200
    assert "Access-Control-Allow-Headers" not in response.headers
    assert response.json() == "pong"


def test_allowed_headers_wildcard(
    client: TestClient,
) -> None:
    client.app.config["cors"]["allowed_headers"] = ["*"]

    response = client.post(
        "/api/ping",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Headers": "X-Custom-3",
        },
    )

    assert response.status_code == 200
    assert "Access-Control-Allow-Headers" not in response.headers
    assert response.json() == "pong"


def test_allowed_headers_post_not_allowed(
    client: TestClient,
) -> None:
    response = client.post(
        "/api/ping",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Headers": "X-Custom-3",
        },
    )

    assert response.status_code == 200
    assert "Access-Control-Allow-Headers" not in response.headers


def test_error(
    client: TestClient,
) -> None:
    response = client.post(
        "/api/error",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Method": "POST",
        },
    )

    assert response.status_code == 500
    assert response.headers["Access-Control-Allow-Origin"] == "http://localhost"


def test_supports_credentials(
    client: TestClient,
) -> None:
    client.app.config["cors"]["supports_credentials"] = True
    response = client.options(
        "/api/ping",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Method": "POST",
        },
    )

    assert response.status_code == 204
    assert response.headers["Access-Control-Allow-Origin"] == "http://localhost"
    assert response.headers["Access-Control-Allow-Credentials"] == "true"


def test_no_matching_path(
    client: TestClient,
) -> None:
    response = client.post(
        "/web/ping",
        headers={"Origin": "http://localhost", "Access-Control-Request-Method": "POST"},
    )

    assert response.status_code == 200
    assert "Access-Control-Allow-Origin" not in response.headers
    assert response.text == "pong"


def test_exposed_headers(
    client: TestClient,
) -> None:
    client.app.config["cors"]["exposed_headers"] = ["X-Foo"]
    response = client.post(
        "/api/ping",
        headers={
            "Origin": "http://localhost",
            "Access-Control-Request-Method": "POST",
        },
    )

    assert response.status_code == 200
    assert response.headers["Access-Control-Allow-Origin"] == "http://localhost"
    assert response.headers["Access-Control-Expose-Headers"] == "X-Foo"



# ruff: noqa: I002
import builtins
import inspect
import logging
import types

from collections.abc import Awaitable
from collections.abc import Callable
from inspect import isgeneratorfunction
from typing import Annotated
from typing import Any
from typing import Self
from typing import TypeVar
from typing import get_args
from typing import get_origin
from typing import overload

from expanse.common.container.container import Container as BaseContainer
from expanse.common.container.exceptions import ContainerException
from expanse.common.container.exceptions import ResolutionException
from expanse.common.container.exceptions import UnboundAbstractException
from expanse.common.support._utils import string_to_class


T = TypeVar("T")
ReturnType = TypeVar("ReturnType")

_builtins = [d for d in dir(builtins) if isinstance(getattr(builtins, d), type)]
_EMPTY = object()

logger = logging.getLogger(__name__)

_Callback = Callable[..., None]


class Container(BaseContainer):
    def __init__(self) -> None:
        super().__init__()

        self._terminating_callbacks: list[_Callback] = []
        self._scoped_terminating_callbacks: list[_Callback] = []

    def build(
        self, concrete: type | str, args: tuple | None = None
    ) -> tuple[Any, _Callback | None]:
        if args is None:
            args = ()

        function: Callable[..., Any]
        if isinstance(concrete, types.FunctionType):
            if concrete.__name__ == "<lambda>":
                return concrete(self, *args), None

            function = concrete
        elif isinstance(concrete, types.MethodType):
            function = concrete
        else:
            if isinstance(concrete, str):
                concrete = string_to_class(concrete)

            if isinstance(concrete, types.MethodType):
                function = concrete
            else:
                # What we are trying to build is a class,
                # so we need to resolve the parameters of the __init__ method.
                function = concrete.__init__  # type: ignore[misc]

        (
            positional,
            keywords,
        ) = self._resolve_callable_dependencies(function, *args)

        if isgeneratorfunction(concrete):
            generator = concrete(*positional, **keywords)

            def terminating_callback() -> None:
                next(generator, None)

            return next(generator), terminating_callback

        return concrete(*positional, **keywords), None

    @overload
    def get(self, abstract: type[T]) -> T: ...

    @overload
    def get(self, abstract: str) -> Any: ...

    def get(self, abstract: str | type[T]) -> Any | T:
        return self._resolve(abstract)

    def call(
        self, callable: Callable[..., ReturnType], *args: Any, **kwargs: Any
    ) -> ReturnType:
        if (
            isinstance(callable, types.FunctionType)
            and "." in callable.__qualname__
            and not inspect.ismethod(callable)
            and "<locals>" not in callable.__qualname__
        ):
            # We have an instance method, so we will retrieve the corresponding class,
            # resolve it and call the method.
            class_name, func_name = callable.__qualname__.rsplit(".", maxsplit=1)
            class_: type = callable.__globals__[class_name]

            instance: Any = self.get(class_)

            callable = getattr(instance, func_name)

        (
            positional,
            keywords,
        ) = self._resolve_callable_dependencies(callable, *args, **kwargs)

        return callable(*positional, **keywords)

    def terminating(self, callback: _Callback, scoped: bool = False) -> None:
        if scoped:
            self._scoped_terminating_callbacks.append(callback)
        else:
            self._terminating_callbacks.append(callback)

    def terminate(self) -> None:
        for callback in self._terminating_callbacks:
            self.call(callback)

    def create_scoped_container(self) -> "ScopedContainer":
        container = ScopedContainer(self)

        return container

    def on_resolved(
        self,
        abstract: str | type,
        callback: _Callback,
    ) -> None:
        if not self._is_scoped(abstract) and self.resolved(abstract):
            callback(self.get(abstract))

        self.after_resolving(abstract, callback)

    def _concrete_closure(
        self, abstract: str | type, concrete: Any
    ) -> Callable[[Self], Awaitable[Any]]:
        original_concrete = concrete

        def closure(container: Container) -> Any:
            if abstract == original_concrete:
                obj, _ = container.build(original_concrete)
                return obj

            return container._resolve(original_concrete)

        return closure

    @overload
    def _resolve(self, abstract: type[T]) -> T: ...

    @overload
    def _resolve(self, abstract: str) -> Any: ...

    def _resolve(self, abstract: str | type[T]) -> Any | T:
        abstract = self._get_alias(abstract)

        if abstract in self._instances:
            return self._instances[abstract]

        metadata: tuple = ()
        actual_abstract: str | type[T] = abstract
        origin = get_origin(abstract)
        if origin is Annotated:
            actual_abstract, *metadata = get_args(abstract)  # type: ignore[assignment]

        if actual_abstract in self._bindings:
            concrete = self._bindings[actual_abstract]["concrete"]
        elif isinstance(abstract, str):
            # Unbound strings cannot be resolved
            raise UnboundAbstractException(
                f"Unbound abstract [{abstract}] cannot be resolved"
            )
        else:
            concrete = abstract

        terminating_callback: _Callback | None = None
        if self._can_build(actual_abstract, concrete):
            try:
                obj, terminating_callback = self.build(concrete, metadata)
            except Exception as e:
                raise ContainerException(
                    f'Unable to build the "{abstract}" dependency'
                ) from e
        else:
            obj = self.get(concrete)

        if self._is_cached(actual_abstract):
            self._instances[abstract] = obj

        self._mark_as_resolved(actual_abstract)
        if terminating_callback is not None:
            self.terminating(
                terminating_callback, scoped=self._is_scoped(actual_abstract)
            )

        self._execute_after_resolving_callbacks(abstract, obj)

        return obj

    def _resolve_callable_dependencies(
        self, callable: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> tuple[list[Any], dict[str, Any]]:
        positional: list[Any] = []
        keywords: dict[str, Any] = {}
        arguments = list(args)
        _globals = getattr(callable, "__globals__", None)

        try:
            signature = inspect.signature(callable)
        except ValueError:
            return positional, keywords

        for name, parameter in signature.parameters.items():
            klass = self._get_class(parameter)

            if name == "self":
                continue

            if klass is None:
                self._resolve_primitive(
                    parameter, arguments, kwargs, positional, keywords
                )
            else:
                try:
                    self._resolve_class(
                        parameter,
                        arguments,
                        kwargs,
                        positional,
                        keywords,
                        _globals=_globals,
                    )
                except ContainerException as e:
                    raise ResolutionException(
                        f'Unable to resolve dependency with name "{parameter.name}" '
                        f'(type: {klass.__module__ + "." + klass.__qualname__}) '
                        f'in "{callable.__module__ + "." + callable.__qualname__}"'
                    ) from e

        return positional, keywords

    def _resolve_primitive(
        self,
        parameter: inspect.Parameter,
        args: list[Any],
        kwargs: dict[str, Any],
        positional: list[Any],
        keywords: dict[str, Any],
    ) -> None:
        match parameter.kind:
            case parameter.POSITIONAL_ONLY:
                if not args:
                    raise ResolutionException(
                        f'Unable to resolve dependency with name "{parameter.name}"'
                    )

                positional.append(args.pop(0))

                return

            case parameter.POSITIONAL_OR_KEYWORD:
                # Check in keyword arguments first
                if parameter.name in kwargs:
                    keywords[parameter.name] = kwargs.pop(parameter.name)

                elif args:
                    positional.append(args.pop(0))

                elif parameter.default is not parameter.empty:
                    keywords[parameter.name] = parameter.default

                return

            case parameter.KEYWORD_ONLY:
                if parameter.name in kwargs:
                    keywords[parameter.name] = kwargs.pop(
                        parameter.name,
                    )
                elif parameter.default is not parameter.empty:
                    keywords[parameter.name] = parameter.default

                return

            case parameter.VAR_KEYWORD:
                keywords.update(kwargs.copy())

                kwargs.clear()

                return

            case parameter.VAR_POSITIONAL:
                positional.extend(args.copy())

                args.clear()

                return

            case _:
                pass

        raise ResolutionException(
            f'Unable to resolve dependency with name "{parameter.name}"'
        )

    def _resolve_class(
        self,
        parameter: inspect.Parameter,
        args: list[Any],
        kwargs: dict[str, Any],
        positional: list[Any],
        keywords: dict[str, Any],
        *,
        _globals: dict[str, Any] | None = None,
    ) -> Any:
        match parameter.kind:
            case parameter.POSITIONAL_ONLY:
                klass = self._get_class(parameter, _globals=_globals)

                assert klass is not None

                if self.has(self._get_alias(klass)):
                    result = self.get(self._get_alias(klass))
                else:
                    try:
                        result = self.get(self._get_alias(klass))
                    except Exception as e:
                        if not args:
                            raise e

                        arg = args[0]

                        if not isinstance(arg, klass):
                            raise e

                        result = args.pop(0)

                positional.append(result)
                return

            case parameter.POSITIONAL_OR_KEYWORD:
                # Check in keyword arguments first
                if parameter.name in kwargs:
                    keywords[parameter.name] = kwargs.pop(parameter.name)
                else:
                    klass = self._get_class(parameter, _globals=_globals)

                    assert klass is not None

                    if self.has(self._get_alias(klass)):
                        result = self.get(self._get_alias(klass))
                    else:
                        try:
                            result = self.get(self._get_alias(klass))
                        except Exception as e:
                            if not args:
                                raise e

                            arg = args[0]

                            if not isinstance(arg, klass):
                                raise e

                            result = args.pop(0)

                    positional.append(result)

                return

            case parameter.KEYWORD_ONLY:
                if parameter.name in kwargs:
                    keywords[parameter.name] = kwargs.pop(
                        parameter.name,
                    )
                else:
                    klass = self._get_class(parameter, _globals=_globals)

                    assert klass is not None

                    result = self.get(self._get_alias(klass))
                    keywords[parameter.name] = result
                return

            case parameter.VAR_POSITIONAL:
                klass = self._get_class(parameter, _globals=_globals)

                assert klass is not None

                result = self.get(self._get_alias(klass))

                result = [result] if not isinstance(result, tuple) else result

                positional.extend(result)
                return

            case _:
                pass

        raise ResolutionException(
            f'Unable to resolve dependency with name "{parameter.name}"'
        )

    def _execute_after_resolving_callbacks(
        self, abstract: str | type, instance: Any
    ) -> None:
        callbacks: list[_Callback] = []
        abstract = self._get_alias(abstract)

        actual_abstract: str | type = abstract
        origin = get_origin(abstract)
        if origin is Annotated:
            actual_abstract, *_ = get_args(abstract)

        if abstract in self._after_resolving_callbacks:
            callbacks += self._after_resolving_callbacks[abstract]
        elif actual_abstract in self._after_resolving_callbacks:
            callbacks += self._after_resolving_callbacks[actual_abstract]

        for callback in callbacks:
            callback(instance)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        self.terminate()


class ScopedContainer(Container):
    def __init__(self, base_container: Container):
        super().__init__()

        self._base_container = base_container

        # Bind scoped bindings from the base container
        self._bindings.update(
            {k: {**v} for k, v in self._base_container._scoped["bindings"].items()}
        )

        # Setup terminating callbacks
        self._terminating_callbacks = [
            *self._base_container._scoped["terminating_callbacks"]
        ]

        # Setup resolving callbacks
        self._after_resolving_callbacks = {
            **self._base_container._scoped["after_resolving_callbacks"]
        }

        # Copy instances from the base container
        self._instances.update(self._base_container._instances)

        self.instance(Container, self)

    def bound(self, abstract: str | type) -> bool:
        return self._base_container.bound(abstract) or super().bound(abstract)

    def _directly_bound(self, abstract: str | type) -> bool:
        return super().bound(abstract)

    def _resolve(self, abstract: str | type[T]) -> Any | T:
        actual_abstract: str | type[T] = abstract
        if hasattr(abstract, "__metadata__"):
            actual_abstract, *_ = get_args(abstract)

        # If the abstract is neither bound in the container nor in its base container,
        # we will resolve it from the scoped container.
        if not self.bound(abstract):
            return super()._resolve(abstract)

        if not self._directly_bound(actual_abstract):
            return self._base_container._resolve(abstract)

        return super()._resolve(abstract)



from typing import TYPE_CHECKING

from expanse.common.configuration.config import Config
from expanse.static.static import Static
from expanse.support.service_provider import ServiceProvider
from expanse.view.view_factory import ViewFactory


if TYPE_CHECKING:
    from pathlib import Path

    from expanse.core.application import Application
    from expanse.routing.router import Router


class StaticServiceProvider(ServiceProvider):
    def register(self) -> None:
        self._container.singleton(Static, self._register_static)

    def boot(self) -> None:
        self._container.on_resolved("router", self._add_static_route)
        self._container.on_resolved("view", self._register_view_globals)

    def _register_static(self, config: Config) -> Static:
        app: Application = self._container.get("app")
        paths: list[Path] = config.get("static.paths", [])
        paths = [app.base_path / p if not p.is_absolute() else p for p in paths]
        if (url := config.get("static.url")) is not None:
            url = str(url)

        return Static(paths, prefix=config.get("static.prefix"), url=url)

    def _add_static_route(self, router: "Router") -> None:
        config = self._container.get(Config)
        if config.get("app.debug", False):
            prefix: str = config["static.prefix"].rstrip("/")

            router.get(
                f"{prefix}/{{path:path}}",
                self._container.get(Static).get,
                name="static",
            )

    def _register_view_globals(self, view: ViewFactory) -> None:
        view.register_global(static=self._container.get(Static).url)



import logging

from pathlib import Path

from _pytest.logging import LogCaptureFixture

from expanse.asynchronous.core.application import Application
from expanse.asynchronous.core.http.middleware.middleware_stack import MiddlewareStack
from expanse.asynchronous.http.request import Request
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.testing.client import TestClient
from expanse.asynchronous.types.http.middleware import RequestHandler


logger = logging.getLogger(__name__)


class Middleware1:
    async def handle(self, request: Request, next_call: RequestHandler) -> Response:
        logger.info("Middleware 1")

        response = await next_call(request)

        response.headers["X-Middleware-1"] = "True"

        return response


class Middleware2:
    async def handle(self, request: Request, next_call: RequestHandler) -> Response:
        logger.info("Middleware 2")

        response = await next_call(request)

        response.headers["X-Middleware-2"] = "True"

        return response


class Middleware3:
    async def handle(self, request: Request, next_call: RequestHandler) -> Response:
        logger.info("Middleware 3")

        response = await next_call(request)

        response.headers["X-Middleware-3"] = "True"

        return response


async def configure_middleware(stack: MiddlewareStack) -> None:
    stack.append(Middleware1, Middleware2, Middleware3)


async def test_middleware_configuration(root: Path, caplog: LogCaptureFixture) -> None:
    app = Application.configure(root).with_middleware(configure_middleware).create()
    await app.bootstrap()

    router: Router = await app.container.get(Router)

    router.get("/", lambda: Response("Hello, World!"))

    client = TestClient(app, raise_server_exceptions=True)

    caplog.clear()
    caplog.set_level(logging.INFO)

    response = client.get("/")

    assert "X-Middleware-1" in response.headers
    assert "X-Middleware-2" in response.headers
    assert "X-Middleware-3" in response.headers

    assert caplog.messages[:3] == ["Middleware 1", "Middleware 2", "Middleware 3"]



"""
Auto Migration

Revision ID: 1234567890
Revises:
Create Date: 2024-09-05 12:34:56
"""

import textwrap

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision: str = "1234567890"
down_revision: str | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "users",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("first_name", sa.String(), nullable=True),
        sa.Column("last_name", sa.String(), nullable=True),
        sa.Column("email", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.execute(
        textwrap.dedent(
            """
            INSERT INTO users (first_name, last_name, email)
            VALUES ('John', 'Doe', 'john@doe.com')
            """
        )
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("users")
    # ### end Alembic commands ###



import secrets

from enum import StrEnum
from typing import Any
from typing import ClassVar


class Cipher(StrEnum):
    AES_256_GCM: str = "aes-256-gcm"


class Encryptor:
    CIPHERS: ClassVar[dict[Cipher, dict[str, Any]]] = {
        Cipher.AES_256_GCM: {"key_length": 32, "iv_length": 16}
    }

    def __init__(self, secret_key: bytes, cipher: Cipher = Cipher.AES_256_GCM) -> None:
        self._secret_key = secret_key
        self._cipher = cipher

    @classmethod
    def generate_random_key(cls, cipher: Cipher = Cipher.AES_256_GCM) -> bytes:
        cipher_config = cls.CIPHERS[cipher]
        key = secrets.token_bytes(cipher_config["key_length"])

        return key



from expanse.routing.router import Router
from expanse.testing.client import TestClient
from expanse.view.view import View
from expanse.view.view_factory import ViewFactory


def simple_view(view: ViewFactory) -> View:
    return view.make("simple")


def test_simple_route_are_properly_registered(
    router: Router, client: TestClient
) -> None:
    router.get("/", simple_view)

    response = client.get("/")

    assert response.headers["Content-Type"] == "text/html; charset=utf-8"
    assert "This is a view" in response.text



from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict


class Config(BaseSettings):
    # The cipher method used for encryption. Currently, only aes-256-gcm is supported.
    cipher: str = "aes-256-gcm"

    model_config = SettingsConfigDict(env_prefix="encryption_")



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Protocol


if TYPE_CHECKING:
    from expanse.http.request import Request
    from expanse.http.response import Response
    from expanse.types.http.middleware import RequestHandler


class Middleware(Protocol):
    def handle(self, request: Request, next_call: RequestHandler) -> Response: ...



import textwrap

from typing import Any

from expanse.http.request import Request
from expanse.http.response import Response
from expanse.routing.router import Router


class Redirect:
    def __init__(self, router: Router, request: Request) -> None:
        self._router: Router = router
        self._request: Request = request

    def to(
        self, url: str, status: int = 302, headers: dict[str, Any] | None = None
    ) -> Response:
        return self._create_response(url, status, headers)

    def to_route(
        self,
        name: str,
        parameters: dict[str, Any] | None = None,
        status: int = 302,
        headers: dict[str, Any] | None = None,
    ) -> Response:
        return self._create_response(
            self._router.route(name, parameters=parameters),
            status=status,
            headers=headers,
        )

    def _create_response(
        self, url: str, status: int = 302, headers: dict[str, Any] | None = None
    ) -> Response:
        content = textwrap.dedent(
            """<!DOCTYPE html>
            <html lang="en">
                <head>
                    <meta charset="UTF-8" />
                    <meta http-equiv="refresh" content="0;url=\'%1$s\'" />

                    <title>Redirecting to {url}</title>
                </head>
                <body>
                    Redirecting to <a href="{url}">{url}</a>.
                </body>
            </html>"""
        ).format(
            url=url.replace("&", "&amp;")
            .replace('"', "&quot;")
            .replace("<", "&lt;")
            .replace(">", "&gt;")
        )
        response = Response(
            content, status_code=status, headers=headers, content_type="text/html"
        )
        response.headers["Location"] = url

        return response



from collections import defaultdict
from collections.abc import Iterator
from typing import ClassVar
from typing import NoReturn

from expanse.asynchronous.http.request import Request
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.route import Route
from expanse.common.core.http.exceptions import HTTPException
from expanse.common.routing.route import Match
from expanse.common.routing.route_matcher import RouteMatcher


class RouteCollection:
    METHODS: ClassVar[list[str]] = [
        "GET",
        "HEAD",
        "POST",
        "PUT",
        "PATCH",
        "DELETE",
        "OPTIONS",
    ]

    def __init__(self) -> None:
        self._routes: list[Route] = []
        self._routes_by_method: dict[str, list[Route]] = defaultdict(list)
        self._route_matcher: RouteMatcher = RouteMatcher()

    def add(self, route: Route) -> None:
        self._routes.append(route)

        for method in route.methods:
            self._routes_by_method[method.upper()].append(route)

    def match(self, request: Request):
        routes = self.get(request.method)

        route = self.get_matching_route(request, routes)

        return self._handle_route(request, route)

    def get(self, method: str | None = None) -> list[Route]:
        if method is None:
            return self._routes

        return self._routes_by_method.get(method.upper(), [])

    def get_matching_route(
        self, request: Request, routes: list[Route], allow_partial: bool = False
    ) -> Route | None:
        for route in routes:
            match = self._route_matcher.match(route, request)
            if match == Match.FULL:
                return route

            if match == Match.PARTIAL and allow_partial:
                return route

        return None

    def _handle_route(self, request: Request, route: Route | None) -> Route | None:
        if route is not None:
            return route

        # No route was found.
        # We will check if there is a route that matches the request but with
        # a different method.
        methods = self._check_for_alternative_methods(request)
        if methods:
            return self._get_route_for_methods(request, methods)

        raise HTTPException(404, f"Route {request.url.path} could not be found")

    def _check_for_alternative_methods(self, request: Request) -> list[str]:
        methods = []
        for method, routes in self._routes_by_method.items():
            if method == request.method:
                continue

            routes = self.get(method)

            if self.get_matching_route(request, routes, allow_partial=True) is not None:
                methods.append(method)
                continue

        return methods

    def _get_route_for_methods(
        self, request: Request, methods: list[str]
    ) -> Route | None:
        if request.method == "OPTIONS":

            async def endpoint() -> Response:
                return Response(
                    "", status_code=200, headers={"Allow": ", ".join(methods)}
                )

            return Route(request.url.path, endpoint, methods=["OPTIONS"])

        self._method_not_allowed(request, methods)

    def _method_not_allowed(self, request: Request, methods: list[str]) -> NoReturn:
        raise HTTPException(
            405,
            f"The method {request.method} is not supported for route {request.url.path}. "
            f"Available methods: {', '.join(methods)}",
            headers={"Allow": ", ".join(methods)},
        )

    def __iter__(self) -> Iterator[Route]:
        return iter(self._routes)



from pathlib import Path

from treat.mock import Mockery  # type: ignore[import-untyped]

from expanse.core.application import Application
from expanse.database.console.commands.make_model import MakeModelCommand  # noqa: F401
from expanse.testing.command_tester import CommandTester


def test_make_model(
    command_tester: CommandTester, app: Application, tmp_path: Path, mockery: Mockery
) -> None:
    mockery.mock(app).should_receive("path").with_("models").and_return(
        tmp_path / "models"
    )
    mockery.mock(app).should_receive("path").with_(
        "models/model.py", relative=True
    ).and_return("models/model.py")
    mockery.mock(app).should_receive("path").with_(
        "models/user.py", relative=True
    ).and_return("models/user.py")

    command = command_tester.command("make model")

    assert command.run("User") == 0

    expected = """
  - Generated file: models/user.py
"""

    assert command.output.fetch() == expected

    assert tmp_path.joinpath("models/user.py").exists()

    base_model_content = """\
from expanse.common.database.orm.model import Model


class User(Model):

    __tablename__: str = "users"
"""

    assert tmp_path.joinpath("models/user.py").read_text() == base_model_content


def test_make_model_with_table_name(
    command_tester: CommandTester, app: Application, tmp_path: Path, mockery: Mockery
) -> None:
    mockery.mock(app).should_receive("path").with_("models").and_return(
        tmp_path / "models"
    )
    mockery.mock(app).should_receive("path").with_(
        "models/model.py", relative=True
    ).and_return("models/model.py")
    mockery.mock(app).should_receive("path").with_(
        "models/user.py", relative=True
    ).and_return("models/user.py")

    command = command_tester.command("make model")

    assert command.run("User --table my_users") == 0

    expected = """
  - Generated file: models/user.py
"""

    assert command.output.fetch() == expected

    assert tmp_path.joinpath("models/user.py").exists()

    base_model_content = """\
from expanse.common.database.orm.model import Model


class User(Model):

    __tablename__: str = "my_users"
"""

    assert tmp_path.joinpath("models/user.py").read_text() == base_model_content



from collections.abc import Iterable
from functools import partial
from typing import Self

from expanse.core.application import Application
from expanse.core.http.middleware.middleware import Middleware
from expanse.http.request import Request
from expanse.http.response import Response
from expanse.routing.pipeline import Pipeline
from expanse.routing.router import Router
from expanse.types import Environ
from expanse.types import StartResponse


class Gateway:
    """
    The gateway is the layer between the WSGI spec/world and Expanse internal
    architecture.
    """

    def __init__(self, app: Application, router: Router) -> None:
        self._app = app
        self._router = router
        self._middleware: list[type[Middleware]] = []
        self._group_middleware: dict[str, list[type[Middleware]]] = {}

    def handle(self, request: Request) -> Response:
        with self._app.container.create_scoped_container() as container:
            container.instance(Request, request)

            return (
                Pipeline(container)
                .use(
                    [
                        container.get(middleware).handle
                        for middleware in self._middleware
                    ]
                )
                .send(request)
                .to(partial(self._router.handle, container))
            )

    def set_middleware(self, middleware: list[type[Middleware]]) -> Self:
        self._middleware = middleware

        return self

    def prepend_middleware(self, middleware: type[Middleware]) -> Self:
        if middleware not in self._middleware:
            self._middleware.insert(0, middleware)

        return self

    def append_middleware(self, middleware: type[Middleware]) -> Self:
        if middleware not in self._middleware:
            self._middleware.append(middleware)

        return self

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        request = Request(environ, start_response)

        response = self.handle(request)

        return self._prepare_response(response, environ, start_response)

    def _prepare_response(
        self, response: Response, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        return response(environ, start_response)



from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.testing.client import TestClient
from expanse.asynchronous.view.view import View
from expanse.asynchronous.view.view_factory import ViewFactory


async def simple_view(view: ViewFactory) -> View:
    return await view.make("globals")


def test_simple_route_are_properly_registered(
    router: Router, client: TestClient
) -> None:
    router.get("/", simple_view)
    router.get("/route/{param1:int}", simple_view, name="some.route")

    response = client.get("/")

    # url global
    assert "/foo/bim/42" in response.text
    # route global
    assert "/route/42" in response.text



from expanse.asynchronous.contracts.debug.exception_renderer import (
    ExceptionRenderer as BaseExceptionRenderer,
)
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.testing.client import TestClient
from expanse.common.core.http.exceptions import HTTPException


class ExceptionRenderer(BaseExceptionRenderer):
    def __init__(self) -> None:
        self.rendered: bool = False

    async def render(self, exception: Exception) -> str:
        self.rendered = True

        return "Rendered"


async def error() -> Response:
    raise Exception("Internal error")


async def forbidden() -> Response:
    raise HTTPException(403, "Forbidden")


def test_unhandled_exceptions_are_displayed_with_debug_information_if_debug_mode(
    router: Router, client: TestClient
) -> None:
    router.get("/error", error)

    with client.handle_exceptions():
        response = client.get("/error")

    assert response.status_code == 500
    assert response.headers["Content-Type"] == "text/html; charset=utf-8"
    assert "Internal error" in response.text


def test_unhandled_exceptions_are_displayed_with_basic_information_if_not_debug_mode(
    router: Router, client: TestClient
) -> None:
    client.app.config["app.debug"] = False

    router.get("/error", error)

    with client.handle_exceptions():
        response = client.get("/error")

    assert response.status_code == 500
    assert response.headers["Content-Type"] == "text/html; charset=utf-8"
    assert "Server Error" in response.text


def test_http_exceptions_are_displayed_via_a_dedicated_html_page(
    router: Router, client: TestClient
) -> None:
    router.get("/forbidden", forbidden)

    with client.handle_exceptions():
        response = client.get("/forbidden")

    assert response.status_code == 403
    assert response.headers["Content-Type"] == "text/html; charset=utf-8"
    assert "Forbidden" in response.text


def test_http_exception_are_displayed_via_renderer_if_configured(
    router, client: TestClient
) -> None:
    renderer = ExceptionRenderer()
    client.app.container.instance(BaseExceptionRenderer, renderer)

    router.get("/error", error)

    with client.handle_exceptions():
        response = client.get("/error")

    assert response.status_code == 500
    assert response.headers["Content-Type"] == "text/html; charset=utf-8"
    assert response.text == "Rendered"
    assert renderer.rendered



from collections.abc import Awaitable
from collections.abc import Callable
from collections.abc import Sequence
from dataclasses import asdict
from dataclasses import is_dataclass
from typing import Annotated
from typing import Any
from typing import Self
from typing import TypeVar
from typing import get_args
from typing import get_origin

from expanse.asynchronous.container.container import Container
from expanse.asynchronous.http.response import Response


_Adapter = Callable[..., Awaitable[Response]] | Callable[..., Response]
_Serializer = Callable[..., Awaitable[dict[str, Any]]]
T = TypeVar("T")


class ResponseAdapter:
    def __init__(self, container: Container) -> None:
        self._container: Container = container
        self._adapters: dict[type, _Adapter] = {
            str: self._adapt_string,
            Sequence: self._adapt_sequence,
        }

    async def adapt(
        self, response: Any, declared_response_type: type | None = None
    ) -> Response:
        adapter = self.adapter(response, declared_response_type=declared_response_type)

        if adapter is None:
            raise ValueError(f"Cannot adapt type {type(response)} to a valid response")

        return await self._container.call(
            adapter, response, expected_type=declared_response_type
        )

    def adapter(
        self, response: Any, declared_response_type: type | None = None
    ) -> _Adapter | None:
        for klass, adapter in self._adapters.items():
            if isinstance(response, klass):
                return adapter

        return self._adapter_with_serializer(response, declared_response_type)

    def register_adapter(self, response_type: type, adapter: _Adapter) -> Self:
        self._adapters[response_type] = adapter

        return self

    def _adapter_with_serializer(
        self, response: Any, declared_response_type: type | None = None
    ) -> _Adapter | None:
        serializer = self._find_serializer(response, declared_response_type)

        if not serializer:
            return None

        async def _adapter(response: Any, **kwargs) -> Response:
            from expanse.asynchronous.routing.responder import Responder

            responder = await self._container.get(Responder)

            return await responder.json(await serializer(response))

        return _adapter

    async def _adapt_string(
        self, response: str, container: Container, **kwargs
    ) -> Response:
        from expanse.asynchronous.routing.responder import Responder

        responder = await container.get(Responder)

        return await responder.json(response)

    async def _adapt_sequence(
        self,
        response: Sequence,
        container: Container,
        *,
        expected_type: type | None = None,
    ) -> Response:
        from expanse.asynchronous.routing.responder import Responder

        if expected_type is not None:
            origin: type | None = get_origin(expected_type)
            if (
                origin is not None
                and origin is not Annotated
                and issubclass(origin, Sequence)
            ):
                expected_type = get_args(expected_type)[0]

        serializer: _Serializer | None = None
        if len(response) > 0:
            serializer = self._find_serializer(response[0], expected_type)

        # Adapt each item in the sequence
        new_response: list[Any] = [
            await serializer(item) if serializer is not None else item
            for item in response
        ]

        responder = await container.get(Responder)

        return await responder.json(new_response)

    def _find_serializer(
        self, obj: Any, type_: type | None = None
    ) -> _Serializer | None:
        serializer: _Serializer | None = None
        if type_ is not None:
            origin = get_origin(type_)

            if origin is Annotated:
                from pydantic import BaseModel

                serialization_model = get_args(type_)[1]

                if issubclass(serialization_model, BaseModel):

                    async def _serializer(model: Any) -> dict[str, Any]:
                        return serialization_model.model_validate(model).model_dump()

                    serializer = _serializer

        if serializer is None and is_dataclass(obj):

            async def _serializer(model: Any) -> dict[str, Any]:
                return asdict(model)

            serializer = _serializer

        return serializer



from typing import Self

from expanse.asynchronous.core.http.middleware.middleware import Middleware
from expanse.asynchronous.types.routing import Endpoint
from expanse.common.routing.route import Route as BaseRoute


class Route(BaseRoute):
    def __init__(
        self,
        path: str,
        endpoint: Endpoint,
        *,
        methods: list[str] | None = None,
        name: str | None = None,
    ) -> None:
        super().__init__(path, endpoint, methods=methods, name=name)

        self._middlewares: list[type[Middleware]] = []

    def get_middleware(self) -> list[type[Middleware]]:
        return self._middlewares

    def middleware(self, *middlewares: type[Middleware]) -> Self:
        self._middlewares.extend(middlewares)

        return self

    def prepend_middleware(self, *middlewares: type[Middleware]) -> Self:
        self._middlewares = list(middlewares) + self._middlewares

        return self



from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from expanse.container.container import Container


class ServiceProvider:
    def __init__(self, container: Container) -> None:
        self._container: Container = container

    def register(self) -> None:
        """
        Register any application services.
        """



import logging

from collections.abc import Callable
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Self
from typing import TypeVar

from cleo.io.outputs.output import Output
from crashtest.inspector import Inspector
from pydantic import ValidationError

from expanse.asynchronous.container.container import Container
from expanse.asynchronous.contracts.debug.exception_handler import (
    ExceptionHandler as ExceptionHandlerContract,
)
from expanse.asynchronous.contracts.debug.exception_renderer import ExceptionRenderer
from expanse.asynchronous.http.request import Request
from expanse.asynchronous.http.response import Response
from expanse.common.configuration.config import Config
from expanse.common.core.http.exceptions import HTTPException


if TYPE_CHECKING:
    from crashtest.frame import Frame

_TError = TypeVar("_TError", bound=Exception)

logger = logging.getLogger(__name__)


class ExceptionHandler(ExceptionHandlerContract):
    def __init__(self, container: Container) -> None:
        self._container = container

        self._dont_report: set[type[Exception]] = {HTTPException, ValidationError}
        self._raise_unhandled_exceptions: bool = False
        self._exception_preparers: dict[type, Callable[[Any], Exception]] = {}

    async def report(self, e: Exception) -> None:
        if not await self.should_report(e):
            return

        await self._report_exception(e)

    async def _report_exception(self, e: Exception) -> None:
        # TODO: Better logging handling
        if self._raise_unhandled_exceptions:
            raise e

        logger.exception(e)

    async def should_report(self, e: Exception) -> bool:
        return not any(isinstance(e, klass) for klass in self._dont_report)

    def ignore(self, exception_class: type[Exception]) -> Self:
        self._dont_report.add(exception_class)

        return self

    def stop_ignoring(self, exception_class: type[Exception]) -> Self:
        if exception_class in self._dont_report:
            self._dont_report.remove(exception_class)

        return self

    async def render(self, request: Request, e: Exception) -> Response:
        e = self._prepare_exception(e)

        if isinstance(e, ValidationError):
            return await self._render_validation_exception(e, request)

        return await self._render_exception_response(request, e)

    async def render_for_console(self, output: Output, e: Exception) -> None:
        from cleo.ui.exception_trace import ExceptionTrace

        trace = ExceptionTrace(e)

        trace.render(output)

    async def _render_exception_response(
        self, request: Request, e: Exception
    ) -> Response:
        if request.expects_json():
            return await self._render_json_response(request, e)

        return await self._render_response(request, e)

    async def _render_json_response(self, request: Request, e: Exception) -> Response:
        from expanse.asynchronous.routing.redirect import Redirect
        from expanse.asynchronous.routing.responder import Responder
        from expanse.asynchronous.routing.router import Router
        from expanse.asynchronous.view.view_factory import ViewFactory

        responder = Responder(
            await self._container.get(ViewFactory),
            Redirect(await self._container.get(Router), request),
        )

        return await responder.json(
            await self._convert_exception_to_dict(e),
            status_code=e.status_code if isinstance(e, HTTPException) else 500,
            indent=2,
            headers=e.headers if isinstance(e, HTTPException) else {},
        )

    def prepare_using(
        self, exception_class: type[_TError], preparer: Callable[[_TError], Exception]
    ) -> None:
        """
        Register a preparer for a specific exception class.

        :param exception_class: The exception class to prepare, i.e. convert to another exception.
        :param preparer: The preparer function to use to convert the given exception type.
        """
        self._exception_preparers[exception_class] = preparer

    def _prepare_exception(self, e: Exception) -> Exception:
        classes = (e.__class__, *e.__class__.__bases__)

        preparer = next(
            (
                preparer
                for klass, preparer in self._exception_preparers.items()
                if klass in classes
            ),
            None,
        )

        if not preparer:
            return e

        return preparer(e)

    async def _render_response(self, request: Request, e: Exception) -> Response:
        config = await self._container.get(Config)
        if not isinstance(e, HTTPException) and config.get("app.debug"):
            if self._container.has(ExceptionRenderer):
                return Response(
                    await (await self._container.get(ExceptionRenderer)).render(e),
                    status_code=500,
                    content_type="text/html",
                )

            return Response(
                await self._render_exception_content(e),
                status_code=500,
                content_type="text/plain",
            )

        if not isinstance(e, HTTPException):
            e = HTTPException(500, str(e))

        return await self._render_http_exception(e)

    async def _render_http_exception(self, e: HTTPException) -> Response:
        await self._register_error_paths()

        if view := (await self._get_http_exception_view(e)):
            from expanse.asynchronous.view.view_factory import ViewFactory

            factory = await self._container.get(ViewFactory)

            response = await factory.render(
                await factory.make(view, {"exception": e}, status_code=e.status_code)
            )

            return response

        return Response(
            await self._render_exception_content(e),
            status_code=e.status_code,
            content_type="text/plain",
            headers=e.headers,
        )

    async def _render_validation_exception(
        self, e: ValidationError, request: Request
    ) -> Response:
        if request.expects_json() or request.is_json():
            content = {"code": "validation_error", "detail": []}

            assert isinstance(content["detail"], list)

            for error in e.errors():
                content["detail"].append(
                    {
                        "loc": error["loc"],
                        "message": error["msg"],
                        "type": error["type"],
                    }
                )

            from expanse.asynchronous.routing.redirect import Redirect
            from expanse.asynchronous.routing.responder import Responder
            from expanse.asynchronous.routing.router import Router
            from expanse.asynchronous.view.view_factory import ViewFactory

            responder = Responder(
                await self._container.get(ViewFactory),
                Redirect(await self._container.get(Router), request),
            )

            return await responder.json(content, status_code=422)

        http_exception = HTTPException(422, str(e))

        return await self._render_http_exception(http_exception)

    async def _get_http_exception_view(self, e: HTTPException) -> str | None:
        view = f"errors/{e.status_code}"

        from expanse.asynchronous.view.view_factory import ViewFactory

        factory = await self._container.get(ViewFactory)

        if not factory.exists(view):
            return None

        return view

    async def _register_error_paths(self) -> None:
        import expanse

        from expanse.asynchronous.view.view_finder import ViewFinder

        (await self._container.get(ViewFinder)).add_paths(
            [Path(expanse.__file__).parent.joinpath("common/exceptions/views")]
        )

    async def _render_exception_content(self, e: Exception) -> str:
        config = await self._container.get(Config)
        if config.get("app.debug", False):
            inspector = Inspector(e)

            message = [f"{inspector.exception_name}: {inspector.exception_message}"]

            if inspector.frames:
                message.extend(
                    [
                        f"in {inspector.frames[-1].filename}",
                        f"at line {inspector.frames[-1].lineno}",
                    ]
                )

            return " ".join(message)

        return e.detail if isinstance(e, HTTPException) else "Server Error"

    async def _convert_exception_to_dict(self, e: Exception) -> dict[str, Any]:
        debug = (await self._container.get(Config)).get("app.debug", False)
        if debug:
            inspector = Inspector(e)

            frame: Frame | None = None
            trace: list[dict[str, Any]] = []
            if inspector.frames:
                frame = inspector.frames[-1]
                trace = [
                    {
                        "file": frame.filename,
                        "line": frame.lineno,
                        "function": frame.function,
                    }
                    for frame in reversed(inspector.frames)
                ]

            response: dict[str, Any] = {
                "message": inspector.exception_message,
                "exception": inspector.exception_name,
            }

            if frame:
                response["file"] = frame.filename
                response["line"] = frame.lineno

            if trace:
                response["traceback"] = trace

            return response

        return {"message": e.detail if isinstance(e, HTTPException) else "Server error"}

    def dont_report(self, *e: type[Exception]) -> Self:
        self._dont_report |= set(e)

        return self

    @contextmanager
    def raise_unhandled_exceptions(
        self, raise_exceptions: bool = True
    ) -> Generator[None, None, None]:
        original_value = self._raise_unhandled_exceptions
        self._raise_unhandled_exceptions = raise_exceptions

        yield

        self._raise_unhandled_exceptions = original_value



from __future__ import annotations

from typing import Any


_NOT_FOUND = object()


class Config:
    def __init__(self, config: dict) -> None:
        self._config = config

    def get(self, key: str | None = None, default: Any = None) -> Any:
        config: dict[str, Any] = self._config

        if not key:
            return config

        parts = key.split(".")
        for part in parts:
            if part not in config:
                return default

            config = config[part]

        return config

    def __setitem__(self, item: Any, value: Any) -> Any:
        config: dict[str, Any] = self._config

        parts = item.split(".")
        count = len(parts)
        for i, part in enumerate(parts):
            if i == count - 1:
                config[part] = value
                return

            if part not in config:
                config[part] = {}

            config = config[part]

    def __getitem__(self, item: Any) -> Any:
        value = self.get(item, default=_NOT_FOUND)

        if value is _NOT_FOUND:
            raise KeyError(item)

        return value

    def __contains__(self, item: Any) -> bool:
        config: dict[str, Any] = self._config

        parts = item.split(".")
        for part in parts:
            if part not in config:
                return False

            config = config[part]

        return True

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self._config!r})"



from typing import Any

from sqlalchemy import URL
from sqlalchemy import event
from sqlalchemy import make_url
from sqlalchemy.orm import sessionmaker
from sqlalchemy.util import immutabledict

from expanse.common.database._utils import create_engine
from expanse.core.application import Application
from expanse.database.config import DatabaseConfig
from expanse.database.config import PostgreSQLConfig
from expanse.database.config import SQLiteConfig
from expanse.database.connection import Connection
from expanse.database.engine import Engine
from expanse.database.session import Session


class DatabaseManager:
    def __init__(self, app: Application) -> None:
        self._app: Application = app
        self._engines: dict[str, Engine] = {}
        self._factories: dict[str, sessionmaker] = {}

    def connection(self, name: str | None = None) -> Connection:
        engine = self.configure_engine(name)

        connection = engine.connect()

        return connection

    def session(self, name: str | None = None) -> Session:
        name = name or self.get_default_connection()

        if name in self._factories:
            return self._factories[name]()

        engine = self.configure_engine(name)
        factory = sessionmaker(engine, class_=Session)

        self._factories[name] = factory

        return self._factories[name]()

    def create_base_engine(self, url: URL, **kwargs) -> Engine:
        engine = create_engine(url, engine_class=Engine, **kwargs)

        return engine

    def configure_engine(self, name: str | None = None) -> Engine:
        name = name or self.get_default_connection()

        if name in self._engines:
            return self._engines[name]

        config = self._configuration(name)

        if not config:
            raise

        self._engines[name] = self._create_engine(config)

        return self._engines[name]

    def _create_engine(self, raw_config: dict[str, Any]) -> Engine:
        config = DatabaseConfig.model_validate(raw_config).root

        match config:
            case SQLiteConfig():
                return self._create_sqlite_engine(config)

            case PostgreSQLConfig():
                return self._create_postgresql_engine(config)

    def _create_sqlite_engine(self, config: SQLiteConfig) -> Engine:
        if config.url is not None:
            url = make_url(str(config.url))
        else:
            if config.database is None:
                raise ValueError("The SQLite database path is not configured.")

            database_path = config.database

            database: str
            if database_path == ":memory:":
                database = database_path
            else:
                if not database_path.is_absolute():
                    database_path = self._app.base_path / database_path

                database_path.parent.mkdir(parents=True, exist_ok=True)

                database = database_path.as_posix()

            url = URL(
                drivername="sqlite",
                database=database,
                host=None,
                port=None,
                username=None,
                password=None,
                query=immutabledict(),
            )

        engine = self.create_base_engine(url)

        if config.foreign_key_constraints:

            @event.listens_for(engine, "connect")
            def set_sqlite_pragma(dbapi_connection, connection_record):
                cursor = dbapi_connection.cursor()
                cursor.execute("PRAGMA foreign_keys=ON")
                cursor.close()

        return engine

    def _create_postgresql_engine(self, config: PostgreSQLConfig) -> Engine:
        if config.url is not None:
            url = make_url(str(config.url))
        else:
            drivername: str = "postgresql"
            if config.dbapi is not None:
                drivername += f"+{config.dbapi}"

            query: dict[str, Any] = {}

            if config.sslmode is not None:
                query["sslmode"] = config.sslmode

            url = URL(
                drivername=drivername,
                host=config.host,
                port=config.port,
                database=config.database,
                username=config.username,
                password=config.password,
                query=immutabledict(query),
            )

        engine = self.create_base_engine(
            url, **config.pool.model_dump(exclude_none=True)
        )

        if config.search_path:

            @event.listens_for(engine, "connect", insert=True)
            def set_search_path(dbapi_connection, connection_record):
                existing_autocommit = dbapi_connection.autocommit
                dbapi_connection.autocommit = True
                cursor = dbapi_connection.cursor()
                cursor.execute(f"SET SESSION search_path='{config.search_path}'")
                cursor.close()
                dbapi_connection.autocommit = existing_autocommit

        return engine

    def get_default_connection(self) -> str:
        return self._app.config.get("database.default")

    def _configuration(self, name: str) -> Any:
        connections = self._app.config.get("database.connections", {})

        if name not in connections:
            raise ValueError(f"The database connection [{name}] not configured.")

        return connections[name]



from typing import Any

import pytest

from expanse.asynchronous.routing.router import Router
from expanse.common.routing.exceptions import NotEnoughURLParameters


@pytest.mark.parametrize(
    "path, parameters, expected",
    [
        ("/foo", {}, "/foo"),
        ("/foo/{bar}", {"bar": 42}, "/foo/42"),
        ("/foo/{bar}/{baz}", {"bar": "bim", "baz": 43}, "/foo/bim/43"),
    ],
)
async def test_url_generation(
    router: Router, path: str, parameters: dict[str, Any], expected: str
) -> None:
    assert await router.url(path, parameters) == expected


@pytest.mark.parametrize(
    "path, parameters, expected",
    [
        (
            "/foo/{bar}",
            {},
            "Not enough parameters for URL /foo/{bar}: missing bar",
        ),
        (
            "/foo/{bar}/{baz}",
            {"baz": 43},
            "Not enough parameters for URL /foo/{bar}/{baz}: missing bar",
        ),
        (
            "/foo/{bar}/{baz}",
            {},
            "Not enough parameters for URL /foo/{bar}/{baz}: missing bar, baz",
        ),
    ],
)
async def test_url_generation_with_missing_parameters(
    router: Router, path: str, parameters: dict[str, Any], expected: str
) -> None:
    with pytest.raises(NotEnoughURLParameters) as e:
        await router.url(path, parameters)

    assert str(e.value) == expected


@pytest.mark.parametrize(
    "path, parameters, expected",
    [
        ("foo", {}, "/foo"),
        ("foo.bar", {"bar": 42}, "/foo/42"),
        ("foo.bar.baz", {"bar": "bim", "baz": "43"}, "/foo/bim/43"),
    ],
)
async def test_route_url_generation(
    router: Router, path: str, parameters: dict[str, Any], expected: str
) -> None:
    router.get("/foo", lambda: "foo", name="foo")
    router.get("/foo/{bar}", lambda: "foo", name="foo.bar")
    router.get("/foo/{bar}/{baz:int}", lambda: "foo", name="foo.bar.baz")
    assert await router.route(path, parameters) == expected



from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.testing.client import TestClient
from expanse.common.core.http.exceptions import HTTPException


async def error() -> Response:
    raise Exception("Internal error")


async def forbidden() -> Response:
    raise HTTPException(403, "Forbidden")


def test_unhandled_exceptions_are_returned_with_debug_information_if_debug_mode(
    router: Router, client: TestClient
) -> None:
    router.get("/error", error)

    with client.handle_exceptions():
        response = client.get("/error", headers=[("Accept", "application/json")])

    assert response.status_code == 500
    data = response.json()
    assert data["message"] == "Internal error"
    assert data["exception"] == "Exception"
    assert data["file"] == str(__file__)
    assert data["line"] == 8
    assert "traceback" in data


def test_unhandled_exceptions_are_returned_with_basic_information_if_not_debug_mode(
    router: Router, client: TestClient
) -> None:
    client.app.config["app.debug"] = False

    router.get("/error", error)

    with client.handle_exceptions():
        response = client.get("/error", headers=[("Accept", "application/json")])

    assert response.status_code == 500
    assert response.json() == {"message": "Server error"}


def test_http_exceptions_are_displayed_as_json(
    router: Router, client: TestClient
) -> None:
    client.app.config["app.debug"] = False

    router.get("/forbidden", forbidden)

    response = client.get("/forbidden", headers=[("Accept", "application/json")])

    assert response.status_code == 403
    assert response.json() == {"message": "Forbidden"}



from expanse.asynchronous.http.helpers import json
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.testing.client import TestClient
from expanse.common.http.form import Form
from tests.asynchronous.integration.http.fixtures.request.models import FooModel


async def create_foo(form: Form) -> Response:
    return await json({"bar": form.fields["bar"].value})


async def create_foo_validated(form: Form[FooModel]) -> Response:
    if form.is_submitted() and form.is_valid():
        assert form.data is not None

        return await json({"bar": form.data.bar})

    return await json({"errors": form.errors, "data": form.data})


def test_simple_form_data_are_not_converted_if_no_validation_model(
    router: Router, client: TestClient
) -> None:
    router.post("/", create_foo)

    response = client.post("/", data={"bar": "42"})

    assert response.json() == {"bar": "42"}


def test_simple_form_data_are_converted_if_validation_model(
    router: Router, client: TestClient
) -> None:
    router.post("/", create_foo_validated)

    response = client.post("/", data={"bar": "42"})

    assert response.json() == {"bar": 42}


def test_validation_errors_are_correctly_handled(
    router: Router, client: TestClient
) -> None:
    router.post("/", create_foo_validated)

    response = client.post("/", data={"bar": "foo"})

    assert response.json() == {
        "data": None,
        "errors": [
            {
                "type": "int_parsing",
                "loc": ["bar"],
                "msg": (
                    "Input should be a valid integer, "
                    "unable to parse string as an integer"
                ),
                "input": "foo",
                "url": "https://errors.pydantic.dev/2.3/v/int_parsing",
            }
        ],
    }



from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict


class Config(BaseSettings):
    paths: list[str] = ["api/*"]
    allowed_methods: list[str] = ["*"]
    allowed_origins: list[str] = ["*"]
    allowed_origins_patterns: list[str] = []
    allowed_headers: list[str] = ["*"]
    exposed_headers: list[str] = ["*"]
    max_age: int = 0
    supports_credentials: bool = False

    model_config = SettingsConfigDict(env_prefix="cors_")



from typing import ClassVar

from cleo.helpers import argument
from cleo.helpers import option
from cleo.io.inputs.argument import Argument
from cleo.io.inputs.option import Option

from expanse.common.database.migration.migrator import Migrator
from expanse.database.console.command import MigrationCommand


class MakeMigrationCommand(MigrationCommand):
    name: str = "make migration"

    description = "Create a new migration file."

    arguments: ClassVar[list[Argument]] = [
        argument("name", "The name of the migration")
    ]
    options: ClassVar[list[Option]] = [
        option("auto", None, "Autogenerate the migration")
    ]

    def handle(self, migrator: Migrator) -> int:
        self.line("")

        migrator.make(
            message=self.argument("name"), auto=self.option("auto"), io=self._io
        )

        return 0



from __future__ import annotations

from typing import Self

from expanse.common.http.accept_header_item import AcceptHeaderItem


class AcceptHeader:
    def __init__(self, items: list[AcceptHeaderItem] | None) -> None:
        self._items: list[AcceptHeaderItem] = items or []
        self._sorted: bool = False

    @classmethod
    def from_string(cls, header: str) -> Self:
        values = [p.strip() for p in header.split(",")]

        return cls(
            [
                AcceptHeaderItem.from_string(value).set_index(i)
                for i, value in enumerate(values)
            ]
        )

    def all(self) -> list[AcceptHeaderItem]:
        self._sort()

        return self._items

    def __str__(self) -> str:
        return ", ".join(str(i) for i in self._items)

    def _sort(self) -> None:
        if self._sorted:
            return

        self._items.sort(key=lambda item: (-item.quality, item.index))
        self._sorted = True



from pathlib import Path

from expanse.core.application import Application
from expanse.database.console.commands.db_migrate import DbMigrateCommand  # noqa: F401
from expanse.database.database_manager import DatabaseManager
from expanse.testing.command_tester import CommandTester


def test_migrate(
    command_tester: CommandTester, app: Application, tmp_path: Path
) -> None:
    app.config["paths"]["database"] = (
        Path(__file__).parent.joinpath("fixtures").relative_to(app.base_path)
    )
    app.config["database"]["default"] = "test"

    command = command_tester.command("db migrate")

    return_code = command.run()
    assert return_code == 0

    expected = """
  - Applying migration 1234567890 (Auto Migration)
  - Applying migration 1234567891 (Foo Migration)
"""

    assert command.output.fetch() == expected

    with app.container.get(DatabaseManager).connection() as connection:
        result = connection.execute("SELECT * FROM users")
        assert result.fetchall() == [(1, "John", "Doe", "john@doe.com", "true")]


def test_migrate_with_step(
    command_tester: CommandTester, app: Application, tmp_path: Path
) -> None:
    app.config["paths"]["database"] = (
        Path(__file__).parent.joinpath("fixtures").relative_to(app.base_path)
    )
    app.config["database"]["default"] = "test"

    command = command_tester.command("db migrate")

    return_code = command.run("--step 1")
    assert return_code == 0

    expected = """
  - Applying migration 1234567890 (Auto Migration)
"""

    assert command.output.fetch() == expected


def test_migrate_with_dry_run_mode(
    command_tester: CommandTester, app: Application, tmp_path: Path
) -> None:
    app.config["paths"]["database"] = (
        Path(__file__).parent.joinpath("fixtures").relative_to(app.base_path)
    )
    app.config["database"]["default"] = "test"

    command = command_tester.command("db migrate")

    return_code = command.run("--dry-run")
    assert return_code == 0

    expected = """
CREATE TABLE alembic_version (
    version_num VARCHAR(32) NOT NULL, 
    CONSTRAINT alembic_version_pkc PRIMARY KEY (version_num)
);

-- Running upgrade  -> 1234567890

CREATE TABLE users (
    id INTEGER NOT NULL, 
    first_name VARCHAR, 
    last_name VARCHAR, 
    email VARCHAR, 
    PRIMARY KEY (id)
);

INSERT INTO users (first_name, last_name, email)
VALUES ('John', 'Doe', 'john@doe.com');

INSERT INTO alembic_version (version_num) VALUES ('1234567890') RETURNING version_num;

-- Running upgrade 1234567890 -> 1234567891

ALTER TABLE users ADD COLUMN active BOOLEAN DEFAULT 'true' NOT NULL;

UPDATE alembic_version SET version_num='1234567891' WHERE alembic_version.version_num = '1234567890';

"""

    assert command.output.fetch() == expected



from typing import Self

from expanse.core.http.middleware.middleware import Middleware


class MiddlewareGroup:
    def __init__(self, middlewares: list[type[Middleware]] | None = None) -> None:
        self._middlewares = middlewares or []

    def append(self, *middleware: type[Middleware]) -> Self:
        """
        Append middleware to the middleware stack.
        """
        self._middlewares.extend(middleware)

        return self

    def prepend(self, *middleware: type[Middleware]) -> Self:
        """
        Prepend middleware to the middleware stack.
        """
        self._middlewares = [*middleware, *self._middlewares]

        return self

    def use(self, middleware: list[type[Middleware]]) -> Self:
        """
        Replace the current middleware with the given middleware.
        """
        self._middlewares = middleware

        return self



from typing import Generic
from typing import TypeVar

from httpx import BaseTransport
from httpx import Client as BaseClient

from expanse.common.core.application import Application


T = TypeVar("T", bound=Application)


class TestClient(BaseClient, Generic[T]):
    def __init__(self, app: T, **kwargs) -> None:
        self.app: T = app

        if "transport" not in kwargs:
            kwargs["transport"] = self.transport

        super().__init__(**kwargs)

    @property
    def transport(self) -> BaseTransport | None:
        return None



import base64

from pathlib import Path
from typing import TYPE_CHECKING

from expanse.asynchronous.core.application import Application
from expanse.asynchronous.support.service_provider import ServiceProvider
from expanse.common.encryption.errors import MissingSecretKeyError


if TYPE_CHECKING:
    from expanse.asynchronous.core.console.gateway import Gateway
    from expanse.common.encryption.encryptor import Encryptor


class EncryptionServiceProvider(ServiceProvider):
    async def register(self) -> None:
        from expanse.common.encryption.encryptor import Encryptor

        self._container.singleton(Encryptor, self._create_encryptor)

    async def boot(self) -> None:
        from expanse.asynchronous.core.console.gateway import Gateway

        await self._container.on_resolved(Gateway, self._register_command_path)

    async def _create_encryptor(self, app: Application) -> "Encryptor":
        from expanse.common.encryption.encryptor import Cipher
        from expanse.common.encryption.encryptor import Encryptor

        secret_key: str = app.config.get("app.secret_key")
        cipher: str = app.config.get("encryption.cipher")

        return Encryptor(
            self._normalize_key(secret_key),
            Cipher(cipher),
        )

    async def _register_command_path(self, gateway: "Gateway") -> None:
        await gateway.load_path(Path(__file__).parent.joinpath("console/commands"))

    def _normalize_key(self, key: str) -> bytes:
        if not key:
            raise MissingSecretKeyError()

        if key.startswith("base64:"):
            return base64.b64decode(key[7:])

        return key.encode()



from expanse.asynchronous.http.request import Request
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.routing.url_generator import URLGenerator


async def test_generator_can_generate_simple_urls(router: Router) -> None:
    url = URLGenerator(router, Request.create("http://example.com"))

    assert await url.to("foo/bar") == "http://example.com/foo/bar"
    assert await url.to("foo/bar", secure=True) == "https://example.com/foo/bar"
    assert (
        await url.to("foo/bar?foo=bar", secure=True)
        == "https://example.com/foo/bar?foo=bar"
    )

    url = URLGenerator(router, Request.create("https://example.com"))
    assert await url.to("foo/bar") == "https://example.com/foo/bar"


async def test_generator_can_generate_route_urls(router: Router) -> None:
    url = URLGenerator(router, Request.create("http://example.com"))

    router.get("/", lambda: "", name="index")
    router.get("/foo/bar", lambda: "", name="foo")
    router.get("/foo/bar/{baz}/boom/{boom}", lambda: "", name="bar")

    assert await url.to("foo/bar") == "http://example.com/foo/bar"
    assert await url.to("foo/bar", secure=True) == "https://example.com/foo/bar"
    assert (
        await url.to("foo/bar?foo=bar", secure=True)
        == "https://example.com/foo/bar?foo=bar"
    )

    assert await url.to_route("index") == "/"
    assert await url.to_route("index", {"foo": "bar"}) == "/?foo=bar"
    assert await url.to_route("index", absolute=True) == "http://example.com/"

    assert await url.to_route("foo") == "/foo/bar"
    assert await url.to_route("foo", {"baz": "boom"}) == "/foo/bar?baz=boom"
    assert await url.to_route("foo", absolute=True) == "http://example.com/foo/bar"

    assert (
        await url.to_route("bar", {"baz": "john", "boom": "doe"})
        == "/foo/bar/john/boom/doe"
    )
    assert (
        await url.to_route("bar", {"baz": "john", "boom": "doe", "another": "param"})
        == "/foo/bar/john/boom/doe?another=param"
    )
    assert (
        await url.to_route("bar", {"baz": "john", "boom": "doe"}, absolute=True)
        == "http://example.com/foo/bar/john/boom/doe"
    )



import logging

from collections.abc import Callable
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Self
from typing import TypeVar

from cleo.io.outputs.output import Output
from crashtest.inspector import Inspector
from pydantic import ValidationError

from expanse.common.configuration.config import Config
from expanse.common.core.http.exceptions import HTTPException
from expanse.container.container import Container
from expanse.contracts.debug.exception_handler import (
    ExceptionHandler as ExceptionHandlerContract,
)
from expanse.contracts.debug.exception_renderer import ExceptionRenderer
from expanse.http.request import Request
from expanse.http.response import Response


if TYPE_CHECKING:
    from crashtest.frame import Frame

_TError = TypeVar("_TError", bound=Exception)

logger = logging.getLogger(__name__)


class ExceptionHandler(ExceptionHandlerContract):
    def __init__(self, container: Container) -> None:
        self._container = container

        self._dont_report: set[type[Exception]] = {HTTPException, ValidationError}
        self._raise_unhandled_exceptions: bool = False
        self._exception_preparers: dict[type, Callable[[Any], Exception]] = {}

    def report(self, e: Exception) -> None:
        if not self.should_report(e):
            return

        self._report_exception(e)

    def _report_exception(self, e: Exception) -> None:
        if self._raise_unhandled_exceptions:
            raise e

        logger.exception(e)

    def should_report(self, e: Exception) -> bool:
        return not any(isinstance(e, klass) for klass in self._dont_report)

    def ignore(self, exception_class: type[Exception]) -> Self:
        self._dont_report.add(exception_class)

        return self

    def stop_ignoring(self, exception_class: type[Exception]) -> Self:
        if exception_class in self._dont_report:
            self._dont_report.remove(exception_class)

        return self

    def render(self, request: Request, e: Exception) -> Response:
        e = self._prepare_exception(e)

        if isinstance(e, ValidationError):
            return self._render_validation_exception(e, request)

        return self._render_exception_response(request, e)

    def render_for_console(self, output: Output, e: Exception) -> None:
        from cleo.ui.exception_trace import ExceptionTrace

        trace = ExceptionTrace(e)

        trace.render(output)

    def prepare_using(
        self, exception_class: type[_TError], preparer: Callable[[_TError], Exception]
    ) -> None:
        self._exception_preparers[exception_class] = preparer

    def _prepare_exception(self, e: Exception) -> Exception:
        classes = (e.__class__, *e.__class__.__bases__)

        preparer = next(
            (
                preparer
                for klass, preparer in self._exception_preparers.items()
                if klass in classes
            ),
            None,
        )

        if not preparer:
            return e

        return preparer(e)

    def _render_exception_response(self, request: Request, e: Exception) -> Response:
        if request.expects_json():
            return self._render_json_response(request, e)

        return self._render_response(request, e)

    def _render_json_response(self, request: Request, e: Exception) -> Response:
        from expanse.routing.responder import Responder

        if self._container.has(Responder):
            responder = self._container.get(Responder)
        else:
            from expanse.routing.redirect import Redirect
            from expanse.routing.router import Router
            from expanse.view.view_factory import ViewFactory

            responder = Responder(
                self._container.get(ViewFactory),
                Redirect(self._container.get(Router), request),
            )

        return responder.json(
            self._convert_exception_to_dict(e),
            status_code=e.status_code if isinstance(e, HTTPException) else 500,
            indent=2,
            headers=e.headers if isinstance(e, HTTPException) else None,
        )

    def _render_response(self, request: Request, e: Exception) -> Response:
        if not isinstance(e, HTTPException) and self._container.get(Config).get(
            "app.debug"
        ):
            if self._container.has(ExceptionRenderer):
                return Response(
                    self._container.get(ExceptionRenderer).render(e),
                    status_code=500,
                    content_type="text/html",
                )

            return Response(
                self._render_exception_content(e),
                status_code=500,
                content_type="text/plain",
            )

        if not isinstance(e, HTTPException):
            e = HTTPException(500, str(e))

        return self._render_http_exception(e)

    def _render_http_exception(self, e: HTTPException) -> Response:
        self._register_error_paths()

        if view := self._get_http_exception_view(e):
            from expanse.view.view_factory import ViewFactory

            factory = self._container.get(ViewFactory)

            response = factory.render(
                factory.make(view, {"exception": e}, status_code=e.status_code)
            )

            return response

        return Response(
            self._render_exception_content(e),
            status_code=e.status_code,
            content_type="text/plain",
            headers=e.headers,
        )

    def _render_validation_exception(
        self, e: ValidationError, request: Request
    ) -> Response:
        if request.expects_json() or request.is_json():
            content: dict[str, Any] = {"code": "validation_error"}
            details: list[dict[str, Any]] = []

            for error in e.errors():
                details.append(
                    {
                        "loc": error["loc"],
                        "message": error["msg"],
                        "type": error["type"],
                    }
                )

            content["detail"] = details

            from expanse.routing.redirect import Redirect
            from expanse.routing.responder import Responder
            from expanse.routing.router import Router
            from expanse.view.view_factory import ViewFactory

            responder = Responder(
                self._container.get(ViewFactory),
                Redirect(self._container.get(Router), request),
            )

            return responder.json(content, status_code=422)

        http_exception = HTTPException(422, str(e))

        return self._render_http_exception(http_exception)

    def _get_http_exception_view(self, e: HTTPException) -> str | None:
        view = f"errors/{e.status_code}"

        from expanse.view.view_factory import ViewFactory

        factory = self._container.get(ViewFactory)

        if not factory.exists(view):
            return None

        return view

    def _render_exception_content(self, e: Exception) -> str:
        config = self._container.get(Config)
        if config.get("app.debug", False):
            inspector = Inspector(e)

            message = [f"{inspector.exception_name}: {inspector.exception_message}"]

            if inspector.frames:
                message.extend(
                    [
                        f"in {inspector.frames[-1].filename}",
                        f"at line {inspector.frames[-1].lineno}",
                    ]
                )

            return " ".join(message)

        return e.detail if isinstance(e, HTTPException) else "Server Error"

    def _convert_exception_to_dict(self, e: Exception) -> dict[str, Any]:
        debug = self._container.get(Config).get("app.debug", False)
        if debug:
            inspector = Inspector(e)
            frame: Frame | None = None
            trace: list[dict[str, Any]] = []
            if inspector.frames:
                frame = inspector.frames[-1]
                trace = [
                    {
                        "file": frame.filename,
                        "line": frame.lineno,
                        "function": frame.function,
                    }
                    for frame in reversed(inspector.frames)
                ]

            response: dict[str, Any] = {
                "message": inspector.exception_message,
                "exception": inspector.exception_name,
            }

            if frame:
                response["file"] = frame.filename
                response["line"] = frame.lineno

            if trace:
                response["traceback"] = trace

            return response

        return {"message": e.detail if isinstance(e, HTTPException) else "Server error"}

    def _register_error_paths(self) -> None:
        import expanse

        from expanse.view.view_finder import ViewFinder

        self._container.get(ViewFinder).add_paths(
            [Path(expanse.__file__).parent.joinpath("common/exceptions/views")]
        )

    def dont_report(self, *e: type[Exception]) -> Self:
        self._dont_report |= set(e)

        return self

    @contextmanager
    def raise_unhandled_exceptions(
        self, raise_exceptions: bool = True
    ) -> Generator[None]:
        original_value = self._raise_unhandled_exceptions
        self._raise_unhandled_exceptions = raise_exceptions

        yield

        self._raise_unhandled_exceptions = original_value



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Literal
from typing import Self

from baize.wsgi.responses import PlainTextResponse
from baize.wsgi.responses import Response as BaizeResponse
from baize.wsgi.responses import SmallResponse as BaizeSmallResponse


if TYPE_CHECKING:
    from collections.abc import Iterable
    from collections.abc import MutableMapping

    from baize.datastructures import Cookie

    from expanse.types import Environ
    from expanse.types import StartResponse


class Response:
    def __init__(
        self,
        content: bytes | str | None = None,
        status_code: int = 200,
        headers: MutableMapping[str, str] | None = None,
        content_type: str | None = None,
        *,
        response: BaizeResponse | None = None,
    ) -> None:
        if response is None:
            response = PlainTextResponse(
                content=content or "",
                status_code=status_code,
                headers=headers,
                media_type=content_type,
            )

        self._response = response
        self._status_code = response.status_code
        self._headers = response.headers
        self._cookies = response.cookies

    @property
    def status_code(self) -> int:
        return self._status_code

    @property
    def headers(self) -> MutableMapping[str, str]:
        return self._response.headers

    @property
    def cookies(self) -> list[Cookie]:
        return self._response.cookies

    @property
    def content_type(self) -> str | None:
        if isinstance(self._response, BaizeSmallResponse):
            return self._response.media_type

        if hasattr(self._response, "content_type"):
            return self._response.content_type

        if "Content-Type" in self._response.headers:
            return self._response.headers["Content-Type"]

        return None

    @property
    def charset(self) -> str | None:
        if isinstance(self._response, BaizeSmallResponse):
            return self._response.charset

        return None

    def with_status(self, status_code: int) -> Self:
        self._response.status_code = status_code

        return self

    def with_header(self, key: str, value: str) -> Self:
        self._response.headers[key] = value

        return self

    def with_headers(self, headers: MutableMapping[str, str]) -> Self:
        self._response.headers.update(headers)

        return self

    def set_cookie(
        self,
        key: str,
        value: str = "",
        max_age: int = -1,
        expires: int | None = None,
        path: str = "/",
        domain: str | None = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["strict", "lax", "none"] = "lax",
    ) -> None:
        self._response.set_cookie(
            key,
            value=value,
            max_age=max_age,
            expires=expires,
            path=path,
            domain=domain,
            secure=secure,
            httponly=httponly,
            samesite=samesite,
        )

    def delete_cookie(
        self,
        key: str,
        value: str = "",
        path: str = "/",
        domain: str | None = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["strict", "lax", "none"] = "lax",
    ) -> None:
        self.set_cookie(
            key,
            value=value,
            path=path,
            domain=domain,
            secure=secure,
            httponly=httponly,
            samesite=samesite,
        )

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        return self._response(environ, start_response)  # type: ignore[arg-type]



from __future__ import annotations

import os
import sys

from typing import TYPE_CHECKING
from typing import cast

from expanse.common.console.console import Console as BaseApplication
from expanse.console.commands.command import Command
from expanse.console.commands.help import HelpCommand
from expanse.console.commands.list import ListCommand


if TYPE_CHECKING:
    from cleo.io.inputs.input import Input
    from cleo.io.io import IO
    from cleo.io.outputs.output import Output

    from expanse.core.application import Application as Expanse


class Console(BaseApplication[Command]):
    def __init__(self, app: Expanse) -> None:
        super().__init__()

        self._app = app
        self._container = app.container

    def run(
        self,
        input: Input | None = None,
        output: Output | None = None,
        error_output: Output | None = None,
    ) -> int:
        try:
            io = self.create_io(input, output, error_output)

            self._configure_io(io)

            try:
                exit_code = self._run(io)
            except BrokenPipeError:
                # If we are piped to another process, it may close early and send a
                # SIGPIPE: https://docs.python.org/3/library/signal.html#note-on-sigpipe
                devnull = os.open(os.devnull, os.O_WRONLY)
                os.dup2(devnull, sys.stdout.fileno())
                exit_code = 0
            except Exception as e:
                if not self._catch_exceptions:
                    raise

                self.render_error(e, io)

                exit_code = 1
                # TODO: Custom error exit codes

        except KeyboardInterrupt:
            exit_code = 1

        if self._auto_exit:
            sys.exit(exit_code)

        return exit_code

    @property
    def default_commands(self) -> list[Command]:
        return [HelpCommand(), ListCommand()]

    def get_help_command(self, command: Command) -> Command:
        help_command: HelpCommand = cast(HelpCommand, self.get("help"))
        help_command.set_command(command)

        return help_command

    def _run(self, io: IO) -> int:
        if io.input.has_parameter_option(["--version", "-V"], True):
            io.write_line(self.long_version)

            return 0

        self._setup_command(io)

        assert self._running_command is not None

        exit_code = self._run_command(self._running_command, io)
        self._running_command = None

        return exit_code

    def _run_command(self, command: Command, io: IO) -> int:
        with self._app.container.create_scoped_container() as container:
            command.set_container(container)

            return command.run(io)



from pathlib import Path

import pendulum

from alembic import util
from treat.mock import Mockery  # type: ignore[import-untyped]

from expanse.common.database.migration.migrator import Migrator
from expanse.core.application import Application
from expanse.database.console.commands.make_migration import (
    MakeMigrationCommand,  # noqa: F401
)
from expanse.testing.command_tester import CommandTester


def test_make_migration(
    command_tester: CommandTester, app: Application, tmp_path: Path, mockery: Mockery
) -> None:
    mockery.mock(util).should_receive("rev_id").and_return("1234567890")

    app.config["paths"]["database"] = tmp_path

    command = command_tester.command("make migration")

    with pendulum.travel_to(pendulum.datetime(2024, 9, 5, 12, 34, 56), freeze=True):
        assert command.run("'Foo Migration'") == 0

    expected = f"""
  - Creating directory {tmp_path}/migrations...
  - Creating directory {tmp_path}/migrations... Done

  - Creating directory {tmp_path}/migrations/versions...
  - Creating directory {tmp_path}/migrations/versions... Done

  - Generating {tmp_path}/migrations/script.py.mako...
  - Generating {tmp_path}/migrations/script.py.mako... Done

  - Generating {tmp_path}/migrations/env.py...
  - Generating {tmp_path}/migrations/env.py... Done

  - Generating {tmp_path}/migrations/alembic.ini...
  - Generating {tmp_path}/migrations/alembic.ini... Done

  - Generating {tmp_path}/migrations/versions/2024_09_05_123456_1234567890_foo_migration.py...
  - Generating {tmp_path}/migrations/versions/2024_09_05_123456_1234567890_foo_migration.py... Done

"""

    assert command.output.fetch() == expected


def test_make_migration_autogenerated(
    command_tester: CommandTester, app: Application, tmp_path: Path, mockery: Mockery
) -> None:
    app.config["paths"]["database"] = tmp_path
    app.config["database"]["default"] = "test"

    migrator = app.container.get(Migrator)
    migrator.config.template_dir = Path(__file__).parent.joinpath("fixtures/templates")

    mockery.mock(util).should_receive("rev_id").and_return("1234567890")
    mockery.mock(app).should_receive("path").with_("models").and_return(
        Path(__file__).parent.joinpath("fixtures/models")
    )

    command = command_tester.command("make migration")

    with pendulum.travel_to(pendulum.datetime(2024, 9, 5, 12, 34, 56), freeze=True):
        return_code = command.run("'Auto Migration' --auto")

    assert return_code == 0

    expected = f"""
  - Creating directory {tmp_path}/migrations...
  - Creating directory {tmp_path}/migrations... Done

  - Creating directory {tmp_path}/migrations/versions...
  - Creating directory {tmp_path}/migrations/versions... Done

  - Generating {tmp_path}/migrations/script.py.mako...
  - Generating {tmp_path}/migrations/script.py.mako... Done

  - Generating {tmp_path}/migrations/env.py...
  - Generating {tmp_path}/migrations/env.py... Done

  - Generating {tmp_path}/migrations/alembic.ini...
  - Generating {tmp_path}/migrations/alembic.ini... Done

  - Generating {tmp_path}/migrations/versions/2024_09_05_123456_1234567890_auto_migration.py...
  - Generating {tmp_path}/migrations/versions/2024_09_05_123456_1234567890_auto_migration.py... Done

"""

    assert command.output.fetch() == expected

    content = tmp_path.joinpath(
        "migrations/versions/2024_09_05_123456_1234567890_auto_migration.py"
    ).read_text()

    assert (
        content
        == """\
\"""
Auto Migration

Revision ID: 1234567890
Revises: 
Create Date: 2024-09-05 12:34:56
\"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '1234567890'
down_revision: str | None = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('users',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('first_name', sa.String(), nullable=True),
    sa.Column('last_name', sa.String(), nullable=True),
    sa.Column('email', sa.String(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('users')
    # ### end Alembic commands ###
"""
    )



import pytest

from expanse.asynchronous.core.application import Application
from expanse.asynchronous.testing.command_tester import CommandTester


@pytest.fixture()
def command_tester(app: Application) -> CommandTester:
    return CommandTester(app)



from __future__ import annotations

from typing import TYPE_CHECKING

from dotenv import load_dotenv


if TYPE_CHECKING:
    from expanse.asynchronous.core.application import Application


class LoadEnvironmentVariables:
    @classmethod
    async def bootstrap(cls, app: Application) -> None:
        env_file = app.environment_path.joinpath(app.environment_file)

        if env_file.exists():
            load_dotenv(env_file)



from typing import TYPE_CHECKING
from typing import ClassVar

from cleo.helpers import argument
from cleo.helpers import option
from cleo.io.inputs.argument import Argument
from cleo.io.inputs.option import Option
from inflection import camelize
from inflection import pluralize
from inflection import underscore

from expanse.common.stubs.generator import Generator
from expanse.core.application import Application
from expanse.database.console.command import MigrationCommand


if TYPE_CHECKING:
    from pathlib import Path


class MakeModelCommand(MigrationCommand):
    name: str = "make model"

    description = "Create a new model file."

    arguments: ClassVar[list[Argument]] = [argument("name", "The name of the model.")]
    options: ClassVar[list[Option]] = [
        option(
            "table",
            None,
            "The name of table associated with the model. Derived from the model name by default.",
            flag=False,
        )
    ]

    def handle(self, app: Application, generator: Generator) -> int:
        self.line("")

        models_path: Path = app.path("models")
        if not models_path.exists():
            models_path.mkdir()
            models_path.joinpath("__init__.py").touch()

        model_name: str = self.argument("name")
        model_name = camelize(model_name)
        table_name: str = self.option("table") or self._infer_table(model_name)
        module_name = underscore(model_name)

        stub = generator.stub("database/model")

        if models_path.joinpath(f"{module_name}.py").exists():
            self.line(
                f'  - <warning>Model file <c1>{app.path(f"models/{module_name}.py", relative=True)}</c1> already exists.</warning>'
            )

            return 1

        stub.generate_to(
            models_path.joinpath(f"{module_name}.py"),
            {
                "model_name": model_name,
                "table_name": table_name,
            },
        )

        self.line(
            f'  - Generated file: <c1>{app.path(f"models/{module_name}.py", relative=True)}</c1>'
        )

        return 0

    def _infer_table(self, model_name: str) -> str:
        return pluralize(underscore(model_name))



import pytest

from sqlalchemy import URL
from sqlalchemy.orm import Session
from sqlalchemy.util import immutabledict
from treat.mock import Mockery  # type: ignore[import-untyped]

from expanse.common.configuration.config import Config
from expanse.common.database._utils import create_engine
from expanse.core.application import Application
from expanse.database import database_manager
from expanse.database.connection import Connection
from expanse.database.database_manager import DatabaseManager
from expanse.database.engine import Engine


@pytest.fixture()
def manager() -> DatabaseManager:
    app = Application()
    app.set_config(
        Config(
            {
                "database": {
                    "default": "sqlite",
                    "connections": {
                        "sqlite": {"driver": "sqlite", "database": ":memory:"},
                        "another_sqlite": {"driver": "sqlite", "url": "sqlite://"},
                        "postgresql": {
                            "driver": "postgresql",
                            "host": "127.0.0.1",
                            "port": 5432,
                            "database": "expanse",
                            "username": "postgres",
                            "password": "password",
                        },
                        "postgresql_psycopg2": {
                            "driver": "postgresql",
                            "dbapi": "psycopg2",
                            "host": "127.0.0.1",
                            "database": "expanse",
                            "username": "postgres",
                            "password": "password",
                            "sslmode": "prefer",
                        },
                        "postgresql_psycopg": {
                            "driver": "postgresql",
                            "url": "postgresql+psycopg://postgres:password@localhost:5432/expanse",
                        },
                        "postgresql_pg8000": {
                            "driver": "postgresql",
                            "url": "postgresql+pg8000://postgres:password@localhost:5432/expanse",
                        },
                    },
                }
            }
        )
    )

    return DatabaseManager(app)


def test_database_manager_can_retrieve_a_default_connection(manager: DatabaseManager):
    assert isinstance(manager.connection(), Connection)


def test_database_manager_can_retrieve_a_default_session(manager: DatabaseManager):
    assert isinstance(manager.session(), Session)


def test_database_manager_can_retrieve_a_named_connection(manager: DatabaseManager):
    assert isinstance(manager.connection("another_sqlite"), Connection)


def test_database_manager_can_retrieve_a_named_session(manager: DatabaseManager):
    assert isinstance(manager.session("another_sqlite"), Session)


@pytest.mark.parametrize(
    "connection_name,expected_url",
    [
        (
            "sqlite",
            URL(
                drivername="sqlite",
                database=":memory:",
                host=None,
                port=None,
                username=None,
                password=None,
                query=immutabledict(),
            ),
        ),
        (
            "another_sqlite",
            URL(
                drivername="sqlite",
                database=None,
                host=None,
                port=None,
                username=None,
                password=None,
                query=immutabledict(),
            ),
        ),
        (
            "postgresql",
            URL(
                drivername="postgresql",
                database="expanse",
                host="127.0.0.1",
                port=5432,
                username="postgres",
                password="password",
                query=immutabledict(),
            ),
        ),
        (
            "postgresql_psycopg2",
            URL(
                drivername="postgresql+psycopg2",
                database="expanse",
                host="127.0.0.1",
                port=5432,
                username="postgres",
                password="password",
                query=immutabledict({"sslmode": "prefer"}),
            ),
        ),
        (
            "postgresql_psycopg",
            URL(
                drivername="postgresql+psycopg",
                database="expanse",
                host="localhost",
                port=5432,
                username="postgres",
                password="password",
                query=immutabledict(),
            ),
        ),
        (
            "postgresql_pg8000",
            URL(
                drivername="postgresql+pg8000",
                database="expanse",
                host="localhost",
                port=5432,
                username="postgres",
                password="password",
                query=immutabledict(),
            ),
        ),
    ],
)
def test_database_manager_builds_a_correct_engine_url(
    manager: DatabaseManager, connection_name: str, expected_url: str, mockery: Mockery
):
    # Create a valid engine first
    engine = create_engine("sqlite://", engine_class=Engine)
    mockery.mock(database_manager).should_receive("create_engine").times(1).with_(
        expected_url, engine_class=Engine
    ).and_return(engine)

    manager.connection(connection_name)



from abc import ABC
from abc import abstractmethod


class ExceptionRenderer(ABC):
    @abstractmethod
    def render(self, exception: Exception) -> str:
        """
        Render a given exception as HTML
        """
        ...



from __future__ import annotations

import pytest


@pytest.fixture()
def environ():
    return {
        "REQUEST_METHOD": "GET",
        "SCRIPT_NAME": "",
        "PATH_INFO": "/",
        "HTTP_HOST": "127.0.0.1:8000",
        "HTTP_ACCEPT_ENCODING": "gzip, deflate",
        "HTTP_ACCEPT": "*/*",
        "HTTP_CONNECTION": "127.0.0.1:8000",
        "HTTP_USER_AGENT": "127.0.0.1:8000",
        "QUERY_STRING": "",
    }



from dataclasses import dataclass
from typing import Annotated

from pydantic import BaseModel
from pydantic import ConfigDict

from expanse.routing.router import Router
from expanse.testing.client import TestClient


@dataclass
class Foo:
    bar: str
    baz: int = 42


class User:
    def __init__(self, first_name: str, last_name: str, email: str) -> None:
        self.first_name = first_name
        self.last_name = last_name
        self.email = email


class UserData(BaseModel):
    first_name: str
    email: str

    model_config = ConfigDict(from_attributes=True)


def serialize_dataclass() -> Foo:
    return Foo("bar")


def serialize_with_pydantic_model() -> Annotated[User, UserData]:
    return User("John", "Doe", "john@doe.com")


def serialize_list_with_pydantic_model() -> list[Annotated[User, UserData]]:
    return [User("John", "Doe", "john@doe.com"), User("Jane", "Doe", "jane@doe.com")]


def test_dataclasses_are_automatically_serialized(
    client: TestClient, router: Router
) -> None:
    router.get("/", serialize_dataclass)

    response = client.get("/")

    assert response.json() == {"bar": "bar", "baz": 42}


def test_objects_can_be_serialized_with_pydantic_models(
    client: TestClient, router: Router
) -> None:
    router.get("/", serialize_with_pydantic_model)

    response = client.get("/")

    assert response.json() == {"first_name": "John", "email": "john@doe.com"}


def test_list_of_objects_can_be_serialized_with_pydantic_models(
    client: TestClient, router: Router
) -> None:
    router.get("/", serialize_list_with_pydantic_model)

    response = client.get("/")

    assert response.json() == [
        {"first_name": "John", "email": "john@doe.com"},
        {"first_name": "Jane", "email": "jane@doe.com"},
    ]



from typing import Self

from expanse.asynchronous.core.http.middleware.middleware import Middleware
from expanse.asynchronous.core.http.middleware.middleware_group import MiddlewareGroup


class MiddlewareStack:
    def __init__(self, middlewares: list[type[Middleware]] | None = None) -> None:
        if middlewares is None:
            middlewares = self.get_default_middleware()

        self._middlewares: list[type[Middleware]] = middlewares
        self._groups: dict[str, MiddlewareGroup] = {}

    @property
    def middleware(self) -> list[type[Middleware]]:
        return self._middlewares

    def append(self, *middleware: type[Middleware]) -> Self:
        """
        Append middleware to the middleware stack.
        """
        self._middlewares.extend(middleware)

        return self

    def prepend(self, *middleware: type[Middleware]) -> Self:
        """
        Prepend middleware to the middleware stack.
        """
        self._middlewares = [*middleware, *self._middlewares]

        return self

    def use(self, middleware: list[type[Middleware]]) -> Self:
        """
        Replace the current middleware with the given middleware.
        """
        self._middlewares = middleware

        return self

    def group(self, name: str) -> MiddlewareGroup:
        """
        Retrieve the middleware group with the given name.

        If does not exist it will be created automatically.
        """
        if name not in self._groups:
            self._groups[name] = MiddlewareGroup()

        return self._groups[name]

    def get_default_middleware(self) -> list[type[Middleware]]:
        from expanse.asynchronous.http.middleware.manage_cors import ManageCors

        return [
            ManageCors,
        ]



# ruff: noqa: I002
import uuid

from abc import ABC
from abc import abstractmethod
from collections.abc import Callable
from typing import Any

from expanse.asynchronous.container.container import Container


class Something: ...


class AnotherThing:
    def __init__(self, value: str) -> None:
        self.value = value


class Abstract(ABC):
    @abstractmethod
    def foo(self) -> str: ...

    @abstractmethod
    def run(
        self, foo: str, bar: str, baz: str | None = None
    ) -> tuple[str, str, str | None]: ...

    @abstractmethod
    def run2(
        self, something: Something, foo: str, bar: str, baz: str | None = None
    ) -> tuple[str, str, str | None]: ...

    @abstractmethod
    def run3(
        self,
        something: Something,
        another_thing: AnotherThing,
        foo: str,
        callback: Callable[[int], int],
        bar: str,
        baz: str | None = None,
    ) -> tuple[str, str, int, str, str | None]: ...


class Concrete(Abstract):
    def foo(self) -> str:
        return f"bar {id(self)}"

    def run(
        self, foo: str, bar: str, baz: str | None = None
    ) -> tuple[str, str, str | None]:
        return foo, bar, baz

    def run2(
        self, something: Something, foo: str, bar: str, baz: str | None = None
    ) -> tuple[str, str, str | None]:
        return foo, bar, baz

    def run3(
        self,
        something: Something,
        another_thing: AnotherThing,
        foo: str,
        callback: Callable[[int], int],
        bar: str,
        baz: str | None = None,
    ) -> tuple[str, str, int, str, str | None]:
        return foo, another_thing.value, callback(3), bar, baz


class Foo:
    def __init__(self, concrete: Abstract) -> None:
        self.concrete = concrete
        self._id = str(uuid.uuid4())

    def get_id(self) -> str:
        return self._id


async def test_singleton_returns_same_instance() -> None:
    """
    Singletons should return the same instance.
    """
    container = Container()
    container.singleton(Abstract, Concrete)

    instance1 = await container.get(Abstract)
    instance2 = await container.get(Abstract)

    assert instance1 is instance2


async def test_concrete_is_resolved_directly() -> None:
    container = Container()

    assert isinstance(await container.get(Concrete), Concrete)


async def test_concrete_is_resolved_directly_and_shared() -> None:
    container = Container()
    container.singleton(Concrete)

    instance1 = await container.get(Concrete)
    instance2 = await container.get(Concrete)

    assert instance1 is instance2


async def test_call_resolves_dependencies() -> None:
    def main(concrete: Abstract, foo: str) -> dict[str, Any]:
        return {"concrete": concrete, "foo": foo}

    container = Container()
    container.singleton(Abstract, Concrete)

    result1 = await container.call(main, "bar")
    result2 = await container.call(main, "baz")

    assert result1["concrete"] == result2["concrete"]
    assert result1["foo"] == "bar"
    assert result2["foo"] == "baz"


async def test_call_async_resolves_dependencies() -> None:
    async def main(concrete: Abstract, foo: str) -> dict[str, Any]:
        return {"concrete": concrete, "foo": foo}

    container = Container()
    container.singleton(Abstract, Concrete)

    result1 = await container.call(main, "bar")
    result2 = await container.call(main, "baz")

    assert result1["concrete"] == result2["concrete"]
    assert result1["foo"] == "bar"
    assert result2["foo"] == "baz"


def test_bound() -> None:
    container = Container()

    assert not container.bound(Abstract)
    container.singleton(Abstract, Concrete)
    assert container.bound(Abstract)


async def test_scoped_dependencies() -> None:
    async def foo(container: Container) -> str:
        return await container.get("scoped")

    container = Container()
    container.instance(Container, container)
    container.scoped("scoped", lambda _: str(uuid.uuid4()))

    assert container.has_scoped_bindings()

    async with container.create_scoped_container() as c1:
        result1 = await c1.get("scoped")
        result2 = await c1.get("scoped")

    async with container.create_scoped_container() as c2:
        result3 = await c2.get("scoped")

    assert result1 == result2
    assert result1 != result3


async def test_scoped_container_can_resolve_base_container_dependencies() -> None:
    container = Container()
    container.instance(Container, container)
    container.register(Abstract, Concrete)
    container.scoped(uuid.UUID, uuid.uuid4)

    scoped = container.create_scoped_container()

    result = await scoped.get(Abstract)

    assert isinstance(result, Concrete)


async def test_call_resolves_dependencies_and_parameters_if_parameters_only() -> None:
    container = Container()
    container.singleton(Abstract, Concrete)

    concrete = await container.get(Abstract)

    result = await container.call(concrete.run, "foo", "bar")
    assert result == ("foo", "bar", None)

    result = await container.call(concrete.run, "foo", "bar", baz="baz")
    assert result == ("foo", "bar", "baz")


async def test_call_resolves_deps_and_params_with_mix_of_deps_and_params() -> None:
    container = Container()
    container.singleton(Abstract, Concrete)
    container.singleton(Something)

    concrete = await container.get(Abstract)

    result = await container.call(concrete.run2, "foo", "bar")
    assert result == ("foo", "bar", None)

    result = await container.call(concrete.run2, "foo", "bar", baz="baz")
    assert result == ("foo", "bar", "baz")


async def test_call_resolves_dependencies_and_parameters_with_any_parameter_type() -> (
    None
):
    container = Container()
    container.singleton(Abstract, Concrete)
    container.singleton(Something)

    def callback(i: int) -> int:
        return i

    concrete = await container.get(Abstract)

    result = await container.call(
        concrete.run3, AnotherThing("Value 1"), "foo", callback, "bar"
    )
    assert result == ("foo", "Value 1", 3, "bar", None)

    result = await container.call(
        concrete.run3, AnotherThing("Value 2"), "foo", callback, "bar", baz="baz"
    )
    assert result == ("foo", "Value 2", 3, "bar", "baz")


async def test_call_can_call_instance_methods() -> None:
    container = Container()
    container.instance(Container, container)
    container.register(Abstract, Concrete)

    id1 = await container.call(Foo.get_id)

    assert isinstance(id1, str)

    id2 = await container.call(Foo.get_id)

    assert isinstance(id2, str)
    assert id1 != id2

    container.singleton(Foo)
    await container.get(Foo)

    assert (await container.call(Foo.get_id)) == (await container.call(Foo.get_id))



import dataclasses

from collections.abc import MutableMapping
from typing import Any
from typing import Generic
from typing import NotRequired
from typing import TypeVar
from typing import cast

from baize.datastructures import FormData
from baize.datastructures import MultiMapping
from pydantic import BaseModel
from pydantic import ValidationError
from typing_extensions import TypedDict


Model = TypeVar("Model", bound=type[BaseModel])


class ErrorDetails(TypedDict):
    type: str
    loc: tuple[int | str, ...]
    msg: str
    input: Any
    ctx: NotRequired[dict[str, Any]]


@dataclasses.dataclass
class Field:
    name: str
    value: Any
    error: str | None = None

    def is_valid(self) -> bool:
        return self.error is not None


T = TypeVar("T", bound=BaseModel)


class Form(Generic[T]):
    data: T | None
    _model: type[T] | None = None

    def __init__(self, data: MutableMapping[str, Any] | FormData | None = None) -> None:
        self._submitted = data
        self.fields: dict[str, Field] = {}
        self.errors: list[ErrorDetails] = []
        self.data: T | None = None
        form_data: MutableMapping[str, Any] | MultiMapping = {}
        if self._submitted is not None:
            form_data = self._submitted

        if self._model:
            for field_name, _field_info in self._model.model_fields.items():
                self.fields[field_name] = Field(
                    name=field_name, value=form_data.get(field_name)
                )

            if self._submitted is not None:
                try:
                    self.data = self._model.model_validate(form_data)
                except ValidationError as e:
                    self.errors = e.errors()

                    for error in self.errors:
                        field = self.fields.get(cast(str, error["loc"][0]))
                        if not field:
                            continue

                        field.error = error["msg"]
        else:
            for name, value in form_data.items():
                self.fields[name] = Field(name=name, value=value)

    def is_valid(self) -> bool:
        return not bool(self.errors)

    def is_submitted(self) -> bool:
        return self._submitted is not None

    def __class_getitem__(cls, item: type[T]) -> type["Form"]:
        klass = type(cls.__name__, (cls,), {"_model": item})

        assert issubclass(klass, Form)

        return klass



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import NoReturn

from expanse.core.helpers import _get_container


if TYPE_CHECKING:
    from collections.abc import Mapping
    from collections.abc import MutableMapping

    from expanse.http.response import Response
    from expanse.routing.redirect import Redirect
    from expanse.routing.responder import Responder


def abort(
    status_code: int,
    message: str | None = None,
    headers: MutableMapping[str, str] | None = None,
) -> NoReturn:
    respond().abort(status_code, message, headers)


def json(
    content: Any = "",
    status_code: int = 200,
    headers: MutableMapping[str, Any] | None = None,
    **kwargs: Any,
) -> Response:
    return respond().json(content, status_code, headers, **kwargs)


def text(
    content: Any = "",
    status_code: int = 200,
    headers: MutableMapping[str, Any] | None = None,
) -> Response:
    return respond().text(content, status_code, headers)


def redirect() -> Redirect:
    return respond().redirect()


def respond() -> Responder:
    container = _get_container()

    from expanse.routing.responder import Responder

    return container.get(Responder)


def response(
    content: str = "",
    status_code: int = 200,
    headers: MutableMapping[str, Any] | None = None,
) -> Response:
    return respond().html(content, status_code, headers)


def view(
    view: str,
    data: Mapping[str, Any] | None = None,
    status_code: int = 200,
    headers: MutableMapping[str, Any] | None = None,
) -> Response:
    return respond().view(view, data, status_code, headers)



from collections import defaultdict
from collections.abc import Iterator
from typing import ClassVar
from typing import NoReturn

from expanse.common.core.http.exceptions import HTTPException
from expanse.common.routing.route import Match
from expanse.common.routing.route_matcher import RouteMatcher
from expanse.http.request import Request
from expanse.http.response import Response
from expanse.routing.route import Route


class RouteCollection:
    METHODS: ClassVar[list[str]] = [
        "GET",
        "HEAD",
        "POST",
        "PUT",
        "PATCH",
        "DELETE",
        "OPTIONS",
    ]

    def __init__(self) -> None:
        self._routes: list[Route] = []
        self._routes_by_method: dict[str, list[Route]] = defaultdict(list)
        self._route_matcher: RouteMatcher = RouteMatcher()

    def add(self, route: Route) -> None:
        self._routes.append(route)

        for method in route.methods:
            self._routes_by_method[method.upper()].append(route)

    def match(self, request: Request):
        routes = self.get(request.method)

        route = self.get_matching_route(request, routes)

        return self._handle_route(request, route)

    def get(self, method: str | None = None) -> list[Route]:
        if method is None:
            return self._routes

        return self._routes_by_method.get(method.upper(), [])

    def get_matching_route(
        self, request: Request, routes: list[Route], allow_partial: bool = False
    ) -> Route | None:
        for route in routes:
            match = self._route_matcher.match(route, request)
            if match == Match.FULL:
                return route

            if match == Match.PARTIAL and allow_partial:
                return route

        return None

    def _handle_route(self, request: Request, route: Route | None) -> Route | None:
        if route is not None:
            return route

        # No route was found.
        # We will check if there is a route that matches the request but with
        # a different method.
        methods = self._check_for_alternative_methods(request)
        if methods:
            return self._get_route_for_methods(request, methods)

        raise HTTPException(404, f"Route {request.url.path} could not be found")

    def _check_for_alternative_methods(self, request: Request) -> list[str]:
        methods = []
        for method, routes in self._routes_by_method.items():
            if method == request.method:
                continue

            routes = self.get(method)

            if self.get_matching_route(request, routes, allow_partial=True) is not None:
                methods.append(method)
                continue

        return methods

    def _get_route_for_methods(
        self, request: Request, methods: list[str]
    ) -> Route | None:
        if request.method == "OPTIONS":

            def endpoint() -> Response:
                return Response(
                    "", status_code=200, headers={"Allow": ", ".join(methods)}
                )

            return Route(request.url.path, endpoint, methods=["OPTIONS"])

        self._method_not_allowed(request, methods)

    def _method_not_allowed(self, request: Request, methods: list[str]) -> NoReturn:
        raise HTTPException(
            405,
            f"The method {request.method} is not supported for route {request.url.path}. "
            f"Available methods: {', '.join(methods)}",
            headers={"Allow": ", ".join(methods)},
        )

    def __iter__(self) -> Iterator[Route]:
        return iter(self._routes)



from collections.abc import Generator
from pathlib import Path
from typing import TYPE_CHECKING

from expanse.common.database.migration.migrator import Migrator
from expanse.contracts.debug.exception_handler import ExceptionHandler
from expanse.core.application import Application
from expanse.database.connection import Connection
from expanse.database.database_manager import DatabaseManager
from expanse.database.session import Session
from expanse.support.service_provider import ServiceProvider


if TYPE_CHECKING:
    from expanse.core.console.gateway import Gateway


class DatabaseServiceProvider(ServiceProvider):
    def register(self) -> None:
        self._container.singleton(DatabaseManager)
        self._container.scoped(Session, self._create_session)
        self._container.scoped(Connection, self._create_connection)
        self._container.singleton(Migrator, self._create_migrator)
        self._container.on_resolved(ExceptionHandler, self._configure_exception_handler)

    def boot(self) -> None:
        from expanse.core.console.gateway import Gateway

        self._container.on_resolved(Gateway, self._register_command_path)

    def _create_connection(
        self, db: DatabaseManager, name: str | None = None
    ) -> Generator[Connection]:
        connection = db.connection(name)

        yield connection

        connection.close()

    def _create_session(
        self, db: DatabaseManager, name: str | None = None
    ) -> Generator[Session]:
        session = db.session(name)

        yield session

        session.close()

    def _register_command_path(self, gateway: "Gateway") -> None:
        gateway.load_path(Path(__file__).parent.joinpath("console/commands"))

    def _create_migrator(self, app: Application, db: DatabaseManager) -> Migrator:
        migrator = Migrator(app)

        migrator.config.attributes["db"] = db

        return migrator

    def _configure_exception_handler(self, handler: ExceptionHandler) -> None:
        from sqlalchemy.exc import NoResultFound

        from expanse.common.core.http.exceptions import HTTPException

        def _no_result_found_handler(e: NoResultFound) -> HTTPException:
            return HTTPException(404, str(e))

        # NoResultFound exceptions should be ignored since they will automatically
        # be converted to 404 responses.
        handler.ignore(NoResultFound)
        handler.prepare_using(NoResultFound, _no_result_found_handler)



import inspect

from pathlib import Path
from typing import TypedDict

from crashtest.inspector import Frame
from crashtest.inspector import Inspector
from jinja2 import UndefinedError


class FrameDict(TypedDict):
    filepath: str
    filename: str
    function: str
    lineno: int
    code: str
    line: str
    highlighted_line: str
    number: int


class FrameCollection(TypedDict):
    is_repeated: bool
    repetitions: int
    frames_count: int
    frames: list[FrameDict]


class Trace(TypedDict):
    exception_name: str
    exception_message: str
    frames: list[FrameCollection]


class ExceptionRendererMixin:
    def build_trace(self, error: Exception) -> Trace:
        if isinstance(error, UndefinedError):
            ...
        inspector = Inspector(error)
        trace = Trace(
            exception_name=inspector.exception_name,
            exception_message=inspector.exception_message,
            frames=[],
        )
        frames: list[FrameCollection] = []
        frames_count = len(inspector.frames)

        i = frames_count
        for collection in reversed(inspector.frames.compact()):
            frame_collection: FrameCollection = {
                "is_repeated": collection.is_repeated(),
                "repetitions": collection.repetitions,
                "frames_count": len(collection),
                "frames": [],
            }
            if collection.is_repeated():
                i -= len(collection) * (collection.repetitions + 1)

            for frame in reversed(collection):
                assert isinstance(frame, Frame)
                filepath = frame.filename
                filename = Path(filepath).name
                frame_collection["frames"].append(
                    {
                        "filepath": filepath,
                        "filename": filename,
                        "function": frame.function,
                        "lineno": frame.lineno,
                        "code": self.highlight_frame(frame),
                        "line": frame.line,
                        "highlighted_line": self.highlight_frame(frame, line_only=True),
                        "number": i,
                    }
                )
                i -= 1

            frames.append(frame_collection)

        trace["frames"] = frames

        return trace

    def highlight_frame(self, frame: Frame, line_only: bool = False) -> str:
        from pygments import highlight
        from pygments.formatters.html import HtmlFormatter
        from pygments.lexers import get_lexer_by_name

        lexer_name = "text"
        if frame.filename:
            filename = Path(frame.filename)

            if filename.suffix in {".jinja2", ".j2"}:
                lexer_name = "html+jinja"
            elif filename.suffix == ".py":
                lexer_name = "python"

        lexer = get_lexer_by_name(lexer_name)

        if line_only:
            return highlight(
                frame.line.strip(),
                lexer,
                HtmlFormatter(wrapcode=True),
            )

        if lexer_name == "html+jinja":
            source = [line + "\n" for line in Path(filename).read_text().splitlines()]
        else:
            module = inspect.getmodule(frame.frame)

            if module is None:
                source = [frame.line or ""]
            else:
                source = inspect.getsourcelines(module)[0]

        lines_before = 5
        lines_after = 5

        start_line = frame.lineno - lines_before

        start_line = max(start_line, 1)
        offset = start_line - 1
        length = lines_after + lines_before + 1

        snippet_lines = source[offset : offset + length]

        # The HTML will not properly display starting lines if they are simply new lines
        # So we discard them and adjust the starting line
        snippet_offset = 0
        for line in snippet_lines:
            if line.strip():
                break

            snippet_offset += 1
            start_line += 1

        code = "".join(snippet_lines[snippet_offset:])
        snippet_lineno = frame.lineno - start_line + 1

        return highlight(
            code,
            lexer,
            HtmlFormatter(
                linenos="inline",
                wrapcode=True,
                hl_lines=[snippet_lineno],
                linenostart=start_line,
            ),
        )



from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.testing.client import TestClient


async def endpoint() -> str:
    return ""


async def test_router_should_respond_with_405_http_error_routes_with_alternative_methods(
    router: Router, client: TestClient
) -> None:
    router.get("/test", endpoint)

    response = client.post("/test")

    assert response.status_code == 405
    assert response.headers["Allow"] == "GET, HEAD"


async def test_router_should_respond_to_options_requests(
    router: Router, client: TestClient
) -> None:
    router.get("/test", endpoint)
    router.post("/test", endpoint)
    router.delete("/test", endpoint)
    router.patch("/test", endpoint)
    router.put("/test", endpoint)

    response = client.options("/test")

    assert response.status_code == 200
    assert response.headers["Allow"] == "GET, HEAD, POST, DELETE, PATCH, PUT"



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from expanse.core.application import Application
from expanse.testing.client import TestClient


if TYPE_CHECKING:
    from collections.abc import Generator

    from expanse.core.http.middleware.middleware_stack import MiddlewareStack
    from expanse.routing.router import Router


@pytest.fixture()
def root() -> Path:
    return Path(__file__).parent.parent.parent


@pytest.fixture()
def app(root: Path) -> Application:
    def configure_middleware(stack: MiddlewareStack) -> None:
        stack.use([])

    application = (
        Application.configure(root).with_middleware(configure_middleware).create()
    )
    application.bootstrap()
    application.config["app.debug"] = True

    return application


@pytest.fixture()
def router(app: Application) -> Router:
    return app.container.get("router")


@pytest.fixture()
def client(app: Application) -> Generator[TestClient]:
    with TestClient(app=app, raise_server_exceptions=True) as client:
        yield client



import os

from pathlib import Path
from typing import Self

from expanse.asynchronous.http.helpers import abort
from expanse.asynchronous.http.helpers import respond
from expanse.asynchronous.http.response import Response


class Static:
    def __init__(
        self, directories: list[Path], prefix: str, url: str | None = None
    ) -> None:
        self._directories = directories
        self._prefix: str = prefix.rstrip("/")
        self._url: str | None = url.rstrip("/") if url is not None else None

    async def get(self, path: str) -> Response:
        full_path = self._find(path)

        if not full_path:
            await abort(404)

        return await (await respond()).file(str(full_path))

    def url(self, path: str) -> str:
        static_url = []

        if self._url is not None:
            static_url.append(self._url)
            if self._prefix.lstrip("/"):
                static_url.append(self._prefix.lstrip("/"))
        elif self._prefix:
            static_url.append(self._prefix)

        static_url.append(path)

        return "/".join(static_url)

    def add_path(self, *paths: Path) -> Self:
        self._directories.extend(paths)

        return self

    def _find(self, path: str) -> Path | None:
        for directory in self._directories:
            directory_path = directory.joinpath(path).resolve()

            if (
                os.path.commonpath([directory_path.as_posix(), directory.as_posix()])
                != directory.as_posix()
            ):
                continue

            if not directory_path.exists():
                continue

            return directory_path



from typing import ClassVar

from cleo.helpers import option
from cleo.io.inputs.argument import Argument
from cleo.io.inputs.option import Option

from expanse.common.database.migration.migrator import Migrator
from expanse.common.database.migration.utils import configure_alembic_loggers
from expanse.database.console.command import MigrationCommand


class DbMigrateCommand(MigrationCommand):
    name: str = "db migrate"

    description = "Upgrade the database to the latest revision."

    arguments: ClassVar[list[Argument]] = []
    options: ClassVar[list[Option]] = [
        option("step", None, "The number of migrations to apply.", flag=False),
        option(
            "dry-run",
            None,
            "Display the SQL queries that would be run, instead of executing them",
        ),
    ]

    help = """
Upgrade the database to the latest revision.

A number of migrations to apply can be specified with the --step option.

<info>db migrate --step 2</info>

This will apply the next two migrations.
"""

    def handle(self, migrator: Migrator) -> int:
        self.line("")

        dry_run: bool = self.option("dry-run")

        revision = "head"
        if self.option("step"):
            revision = f'+{self.option("step").removeprefix("+")}'

        if dry_run:
            configure_alembic_loggers(self._io, disable=True)

        migrator.migrate(revision=revision, dry_run=self.option("dry-run"), io=self._io)

        return 0



from __future__ import annotations

import contextlib
import sys

from io import BytesIO
from typing import TYPE_CHECKING
from typing import Any

import httpx

from httpx import Request
from httpx import Response
from httpx._transports.wsgi import WSGIByteStream

from expanse.common.testing.client import TestClient as BaseTestClient
from expanse.contracts.debug.exception_handler import ExceptionHandler
from expanse.core.application import Application
from expanse.exceptions.handler import ExceptionHandler as ConcreteExceptionHandler


if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Generator

    from _typeshed import OptExcInfo


class WSGITransport(httpx.WSGITransport):
    def handle_request(self, request: Request) -> Response:
        request.read()
        wsgi_input = BytesIO(request.content)

        port = request.url.port or {"http": 80, "https": 443}[request.url.scheme]
        environ = {
            "wsgi.version": (1, 0),
            "wsgi.url_scheme": request.url.scheme,
            "wsgi.input": wsgi_input,
            "wsgi.errors": self.wsgi_errors or sys.stderr,
            "wsgi.multithread": True,
            "wsgi.multiprocess": False,
            "wsgi.run_once": False,
            "REQUEST_METHOD": request.method,
            "SCRIPT_NAME": self.script_name,
            "PATH_INFO": request.url.path,
            "QUERY_STRING": request.url.query.decode("ascii"),
            "SERVER_NAME": request.url.host,
            "SERVER_PORT": str(port),
            "SERVER_PROTOCOL": "HTTP/1.1",
            "REMOTE_ADDR": self.remote_addr,
        }
        for header_key, header_value in request.headers.raw:
            key = header_key.decode("ascii").upper().replace("-", "_")
            if key not in ("CONTENT_TYPE", "CONTENT_LENGTH"):
                key = "HTTP_" + key
            environ[key] = header_value.decode("ascii")

        seen_status = None
        seen_response_headers = None
        seen_exc_info = None

        def start_response(
            status: str,
            response_headers: list[tuple[str, str]],
            exc_info: OptExcInfo | None = None,
        ) -> Callable[[bytes], Any]:
            nonlocal seen_status, seen_response_headers, seen_exc_info
            seen_status = status
            seen_response_headers = response_headers
            seen_exc_info = exc_info
            return lambda _: None

        assert isinstance(self.app, Application)

        if not self.app.container.has(ExceptionHandler):
            self.app.container.singleton(ExceptionHandler, ConcreteExceptionHandler)

        handler: ExceptionHandler = self.app.container.get(ExceptionHandler)
        with handler.raise_unhandled_exceptions(self.raise_app_exceptions):
            result = self.app(environ, start_response)

        stream = WSGIByteStream(result)

        assert seen_status is not None
        assert seen_response_headers is not None
        if seen_exc_info and seen_exc_info[0] and self.raise_app_exceptions:
            raise seen_exc_info[1]

        status_code = int(seen_status.split()[0])
        headers = [
            (key.encode("ascii"), value.encode("ascii"))
            for key, value in seen_response_headers
        ]

        return Response(status_code, headers=headers, stream=stream)


class TestClient(BaseTestClient[Application]):
    def __init__(
        self,
        app: Application,
        base_url: str = "http://testserver",
        cookies: httpx._client.CookieTypes = None,
        headers: dict[str, str] | None = None,
        raise_server_exceptions: bool = False,
    ) -> None:
        if headers is None:
            headers = {}

        self._transport: WSGITransport | None = None
        self.raise_server_exceptions = raise_server_exceptions

        headers.setdefault("user-agent", "testclient")
        super().__init__(
            app=app,
            base_url=base_url,
            headers=headers,
            follow_redirects=True,
            cookies=cookies,
        )

    @property
    def transport(self) -> WSGITransport:
        if self._transport is None:
            self._transport = WSGITransport(
                app=self.app, raise_app_exceptions=self.raise_server_exceptions
            )

        return self._transport

    @contextlib.contextmanager
    def handle_exceptions(self, handle_exceptions: bool = True) -> Generator[None]:
        raise_app_exceptions = self.transport.raise_app_exceptions
        self.transport.raise_app_exceptions = not handle_exceptions

        yield

        self.transport.raise_app_exceptions = raise_app_exceptions



from typing import Annotated

import pytest

from expanse.asynchronous.database.session import Session
from expanse.asynchronous.http.helpers import json
from expanse.asynchronous.http.response import Response
from expanse.asynchronous.routing.router import Router
from expanse.asynchronous.testing.client import TestClient


async def default_session(session: Session) -> Response:
    return await json(
        (
            await session.execute("SELECT id FROM my_table WHERE id = 'sqlite' LIMIT 1")
        ).scalar()
    )


def test_the_default_session_can_be_injected(
    router: Router, client: TestClient
) -> None:
    router.get("/default", default_session)

    response = client.get("/default")

    assert response.json() == "sqlite"


@pytest.mark.parametrize(
    "name",
    ["sqlite", "sqlite2", "postgresql", "postgresql_psycopg", "postgresql_asyncpg"],
)
def test_a_named_session_can_be_injected(
    router: Router, client: TestClient, name: str
) -> None:
    async def named_session(connection: Annotated[Session, name]) -> Response:
        return await json(
            await connection.scalar(
                "SELECT id FROM my_table WHERE id = :id LIMIT 1", {"id": name}
            )
        )

    router.get("/named", named_session)

    response = client.get("/named")

    assert response.json() == name



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Self

from expanse.container.container import Container
from expanse.core.http.middleware.middleware_stack import MiddlewareStack


if TYPE_CHECKING:
    from collections.abc import Callable

    from expanse.console.commands.command import Command
    from expanse.core.application import Application
    from expanse.core.console.gateway import Gateway as ConsoleGateway
    from expanse.core.http.gateway import Gateway as HTTPGateway


class ApplicationBuilder:
    def __init__(self, base_path: Path) -> None:
        self._base_path: Path = base_path
        self._container = Container()
        self._register_commands_callback: Callable[[ConsoleGateway], None] | None = None
        self._configure_middleware_stack: Callable[[HTTPGateway], None] | None = None
        self._configure_kernels: Callable[[Container], None] | None = None

    def with_kernels(self) -> Self:
        def configure_kernels(container: Container) -> None:
            from expanse.core.console.gateway import Gateway as ConsoleGateway

            container.singleton(ConsoleGateway)

        self._configure_kernels = configure_kernels

        return self

    def with_commands(self, commands: list[type[Command] | Path] | None = None) -> Self:
        def _register_commands(kernel: ConsoleGateway) -> None:
            nonlocal commands

            if not commands:
                commands = [kernel._app.path("console/commands")]

            command_paths: list[Path] = []
            command_classes: list[type[Command]] = []
            for command in commands:
                if isinstance(command, Path):
                    command_paths.append(command)
                    continue

                command_classes.append(command)

            kernel.add_command_paths(command_paths)
            kernel.add_commands(command_classes)

        self._register_commands_callback = _register_commands

        return self

    def with_middleware(self, callback: Callable[[MiddlewareStack], None]) -> Self:
        def configure_middleware(gateway: HTTPGateway) -> None:
            stack = MiddlewareStack()

            callback(stack)

            gateway.set_middleware(stack.middleware)

        self._configure_middleware_stack = configure_middleware

        return self

    def create(self) -> Application:
        from expanse.core.application import Application

        app = Application(self._base_path, container=self._container)
        if self._configure_kernels is not None:
            self._configure_kernels(self._container)

        if self._register_commands_callback is not None:

            def _register_commands(container: Container) -> None:
                from expanse.core.console.gateway import Gateway as ConsoleGateway

                assert self._register_commands_callback is not None

                container.on_resolved(ConsoleGateway, self._register_commands_callback)

            app.bootstrapping(_register_commands)

        if self._configure_middleware_stack:

            def _configure_middleware(container: Container) -> None:
                from expanse.core.http.gateway import Gateway

                assert self._configure_middleware_stack is not None

                container.on_resolved(Gateway, self._configure_middleware_stack)

            app.bootstrapping(_configure_middleware)

        return app



from typing import Any
from typing import TypeVar
from typing import overload

from sqlalchemy import CursorResult
from sqlalchemy import Executable
from sqlalchemy import Result
from sqlalchemy import ScalarResult
from sqlalchemy import UpdateBase
from sqlalchemy import text
from sqlalchemy import util
from sqlalchemy.engine.interfaces import _CoreAnyExecuteParams
from sqlalchemy.engine.interfaces import _CoreSingleExecuteParams
from sqlalchemy.orm import Session as BaseSession
from sqlalchemy.orm._typing import OrmExecuteOptionsParameter
from sqlalchemy.orm.session import _BindArguments
from sqlalchemy.sql.selectable import TypedReturnsRows


_T = TypeVar("_T", bound=Any)


class Session(BaseSession):
    @overload  # type: ignore[override]
    def execute(
        self,
        statement: UpdateBase,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        _parent_execute_state: Any | None = None,
        _add_event: Any | None = None,
    ) -> CursorResult[Any]: ...

    @overload
    def execute(
        self,
        statement: Executable,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        _parent_execute_state: Any | None = None,
        _add_event: Any | None = None,
    ) -> Result[Any]: ...

    @overload
    def execute(
        self,
        statement: str,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: dict[str, Any] | None = None,
        _parent_execute_state: Any | None = None,
        _add_event: Any | None = None,
    ) -> Result[Any]: ...

    def execute(
        self,
        statement: UpdateBase | Executable | str,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        _parent_execute_state: Any | None = None,
        _add_event: Any | None = None,
    ) -> Result[Any]:
        if isinstance(statement, str):
            statement = text(statement)

        return super().execute(
            statement,
            params,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
            _parent_execute_state=_parent_execute_state,
            _add_event=_add_event,
        )

    @overload
    def scalar(
        self,
        statement: TypedReturnsRows[tuple[_T]],
        params: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> _T | None: ...

    @overload
    def scalar(
        self,
        statement: Executable,
        params: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> Any: ...

    @overload
    def scalar(
        self,
        statement: str,
        params: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> Any: ...

    def scalar(
        self,
        statement: TypedReturnsRows[tuple[_T]] | Executable | str,
        params: _CoreSingleExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> Any:
        if isinstance(statement, str):
            statement = text(statement)

        return super().scalar(
            statement,
            params,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
            **kw,
        )

    @overload
    def scalars(
        self,
        statement: TypedReturnsRows[tuple[_T]],
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> ScalarResult[_T]: ...

    @overload
    def scalars(
        self,
        statement: Executable,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: _BindArguments | None = None,
        **kw: Any,
    ) -> ScalarResult[Any]: ...

    @overload
    def scalars(
        self,
        statement: str,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: dict[str, Any] | None = None,
        **kw: Any,
    ) -> ScalarResult[Any]: ...

    def scalars(
        self,
        statement: Executable | str,
        params: _CoreAnyExecuteParams | None = None,
        *,
        execution_options: OrmExecuteOptionsParameter = util.EMPTY_DICT,
        bind_arguments: dict[str, Any] | None = None,
        **kw: Any,
    ) -> ScalarResult[Any]:
        if isinstance(statement, str):
            statement = text(statement)

        return super().scalars(
            statement,
            params,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
            **kw,
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Self


if TYPE_CHECKING:
    from pathlib import Path

    from expanse.common.configuration.config import Config


class Application:
    def __init__(self, base_path: Path) -> None:
        super().__init__()

        self._base_path: Path = base_path
        self._config_path: Path | None = None
        self._resources_path: Path | None = None
        self._static_path: Path | None = None
        self._environment_path: Path | None = None
        self._database_path: Path | None = None

        self._booted: bool = False
        self._has_been_bootstrapped: bool = False
        self._config: Config

    @property
    def base_path(self) -> Path:
        return self._base_path

    @property
    def config_path(self) -> Path:
        return self._config_path or self._base_path.joinpath("config")

    @property
    def environment_path(self) -> Path:
        return self._environment_path or self._base_path

    def path(self, path: str | Path, relative: bool = False) -> Path:
        path = self._base_path.joinpath("app").joinpath(path)

        if relative:
            path = path.relative_to(self._base_path)

        return path

    def named_path(self, name: str) -> Path:
        path_key = f"paths.{name}"
        if path_key not in self._config:
            raise ValueError(f"Named path {name} is not configured.")

        path: Path = self._config[path_key]

        return self._base_path.joinpath(path)

    @property
    def environment_file(self) -> str:
        return ".env"

    @property
    def config(self) -> Config:
        return self._config

    def set_config(self, config: Config) -> None:
        self._config = config

    def is_booted(self) -> bool:
        return self._booted

    def has_been_bootstrapped(self) -> bool:
        return self._has_been_bootstrapped

    def set_base_path(self, base_path: Path) -> Self:
        self._base_path = base_path

        return self



from collections.abc import Callable
from importlib import import_module
from typing import ClassVar

from expanse.console.commands.command import Command
from expanse.core.console.gateway import Gateway as ConsoleKernel
from expanse.support.service_provider import ServiceProvider


class CommandServiceProvider(ServiceProvider):
    COMMANDS: ClassVar[list[str]] = ["routes list", "serve"]

    def register(self) -> None:
        # TODO: register commands only on starting event once implemented

        console = self._container.get(ConsoleKernel).console

        for command in self.COMMANDS:
            console.command_loader.register_factory(
                command, self._load_command(command)
            )

    def _load_command(self, name: str) -> Callable[[], Command]:
        def _load() -> Command:
            words = name.split(" ")
            module = import_module("expanse.core.console.commands." + ".".join(words))
            command_class = getattr(
                module, "".join(c.title() for c in words) + "Command"
            )
            command: Command = command_class()
            return command

        return _load



from expanse.common.http.query import Query
from expanse.http.response import Response
from expanse.routing.responder import Responder
from expanse.routing.router import Router
from expanse.testing.client import TestClient
from tests.synchronous.integration.http.fixtures.request.models import FooModel


def index_validated(responder: Responder, query: Query[FooModel]) -> Response:
    return responder.json({"bar": query.bar})


def test_simple_form_data_are_converted_with_validation_model(
    router: Router, client: TestClient
) -> None:
    router.post("/", index_validated)

    response = client.post("/?bar=42", data={"bar": "42"})

    assert response.json() == {"bar": 42}



from __future__ import annotations

import re

from typing import Self


class AcceptHeaderItem:
    def __init__(self, value: str, attributes: dict[str, str] | None = None) -> None:
        self._value: str = value
        self._attributes: dict[str, str] = {}
        self._quality: float = 1.0
        self._index: int = 0

        attributes = attributes or {}
        for name, value in attributes.items():
            self.set_attribute(name, value)

    @classmethod
    def from_string(cls, item: str) -> Self:
        parts = item.split(";", maxsplit=1)
        if len(parts) == 1:
            return cls(item)

        matches = re.findall(r"([^;=]+)=([^;=]+)", parts[1])
        attributes = {}

        for match in matches:
            attributes[match[0]] = match[1]

        return cls(parts[0], attributes)

    @property
    def value(self) -> str:
        return self._value

    @property
    def quality(self) -> float:
        return self._quality

    @property
    def index(self) -> int:
        return self._index

    def set_index(self, index: int) -> Self:
        self._index = index

        return self

    def set_attribute(self, name: str, value: str) -> Self:
        if name == "q":
            self._quality = float(value)
        else:
            self._attributes[name] = value

        return self

    def __str__(self) -> str:
        string = self._value

        if self._quality < 1.0:
            string += f";q={self._quality}"

        if self._attributes:
            string += f"; {';'.join(n + ' ' + v for n, v in self._attributes.items())}"

        return string



from collections.abc import Awaitable
from collections.abc import Callable

from expanse.asynchronous.http.request import Request
from expanse.asynchronous.http.response import Response


RequestHandler = Callable[[Request], Awaitable[Response]]



from pathlib import Path

from pydantic import Field
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict

from expanse.database.config import DatabaseConfig
from expanse.database.config import SQLiteConfig


class Config(BaseSettings):
    # Default database connection
    #
    # The default database connection that should be used
    # when no connection is explicitly specified.
    default: str = Field(validation_alias="db_connection", default="sqlite")

    # Database connections
    #
    # The database connections that are defined for your applications.
    # They can all be defined with environment variables in you `.env` file.
    # For instance:
    # >>> DB_CONNECTIONS__SQLITE__DRIVER=sqlite
    # >>> DB_CONNECTIONS__SQLITE__DATABASE=database/database.sqlite

    connections: dict[str, DatabaseConfig] = Field(
        default={"sqlite": SQLiteConfig(database=Path("database/database.sqlite"))}
    )

    model_config = SettingsConfigDict(env_prefix="DB_", env_nested_delimiter="__")



from expanse.contracts.debug.exception_renderer import (
    ExceptionRenderer as ExceptionRendererContract,
)
from expanse.exceptions.exception_renderer import ExceptionRenderer
from expanse.support.service_provider import ServiceProvider


class ExceptionServiceProvider(ServiceProvider):
    def register(self) -> None:
        self._container.singleton(ExceptionRendererContract, ExceptionRenderer)



from pathlib import Path

from pytest_mock import MockerFixture

from expanse.core.application import Application
from expanse.exceptions.exception_renderer import ExceptionRenderer
from expanse.view.view_factory import ViewFactory
from expanse.view.view_finder import ViewFinder
from tests.synchronous.exceptions.fixtures.exceptions import foo


def test_the_correct_stack_trace_representation_is_passed_to_view(
    app: Application, mocker: MockerFixture
) -> None:
    view = app.container.get(ViewFactory)

    renderer = ExceptionRenderer(view, app.container.get(ViewFinder))

    make = mocker.spy(view, "make")

    try:
        foo()
    except Exception as e:
        renderer.render(e)

    call_args = make.call_args_list[0][0]
    assert call_args[0] == "__expanse__/trace"
    assert call_args[1]["error"] == {"type": "Exception", "message": "Custom exception"}

    trace = call_args[1]["trace"]
    assert not trace[0]["is_repeated"]
    assert trace[0]["repetitions"] == -1
    assert trace[0]["frames_count"] == 2

    frames = trace[0]["frames"]
    assert frames[0]["filepath"] == str(
        Path(__file__).parent.joinpath("fixtures/exceptions.py")
    )
    assert frames[0]["filename"] == "exceptions.py"
    assert frames[0]["function"] == "foo"
    assert frames[0]["lineno"] == 7
    assert frames[0]["number"] == 3

    assert frames[1]["filepath"] == str(Path(__file__))
    assert frames[1]["filename"] == "test_exception_renderer.py"
    assert (
        frames[1]["function"]
        == "test_the_correct_stack_trace_representation_is_passed_to_view"
    )
    assert frames[1]["lineno"] == 22
    assert frames[1]["number"] == 2



from expanse.http.helpers import respond
from expanse.http.response_adapter import ResponseAdapter
from expanse.routing.router import Router
from expanse.testing.client import TestClient


class CustomResponseType:
    def __init__(self, content: str) -> None:
        self.content = content


def text_response() -> str:
    return "foo"


def custom_response() -> CustomResponseType:
    return CustomResponseType("Custom response")


def configure_adapter(adapter: ResponseAdapter) -> None:
    adapter.register_adapter(
        CustomResponseType, lambda response: respond().text(response.content)
    )


def test_string_adapter_adapts_response(router: Router, client: TestClient) -> None:
    router.get("/", text_response)

    response = client.get("/")

    assert response.headers["Content-Type"] == "application/json"
    assert response.json() == "foo"


def test_register_new_adapter(router: Router, client: TestClient) -> None:
    client.app.container.on_resolved(ResponseAdapter, configure_adapter)

    router.get("/", custom_response)

    response = client.get("/")

    assert response.headers["Content-Type"] == "text/plain; charset=utf-8"
    assert response.text == "Custom response"



from pathlib import Path

import expanse.common.exceptions

from expanse.common.exceptions.exception_renderer import ExceptionRendererMixin
from expanse.contracts.debug.exception_renderer import (
    ExceptionRenderer as ExceptionRendererContract,
)
from expanse.view.view_factory import ViewFactory
from expanse.view.view_finder import ViewFinder


class ExceptionRenderer(ExceptionRendererContract, ExceptionRendererMixin):
    def __init__(self, view: ViewFactory, finder: ViewFinder) -> None:
        self._view = view

        finder.add_paths(
            [Path(expanse.common.exceptions.__file__).parent.joinpath("views")]
        )

    def render(self, exception: Exception) -> str:
        trace = self.build_trace(exception)

        view = self._view.make(
            "__expanse__/trace",
            {
                "error": {
                    "type": trace["exception_name"],
                    "message": trace["exception_message"],
                },
                "trace": trace["frames"],
                "asset_content": self.asset_content,
            },
        )

        return self._view.render(view, raw=True)

    def asset_content(self, asset_name: str) -> str:
        return (
            Path(expanse.common.exceptions.__file__)
            .parent.joinpath("assets")
            .joinpath(asset_name)
            .read_text()
        )
